# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import pytest

from plopp import Node
from plopp.backends.matplotlib.interactive import InteractiveFig
from plopp.backends.matplotlib.static import StaticFig
from plopp.backends.matplotlib.utils import copy_figure
from plopp.data.testing import data_array
from plopp.graphics.lineview import LineView


def do_test_copy(Fig):
    da = data_array(ndim=1)
    original = Fig(LineView, Node(da))
    copy = copy_figure(original)
    assert original.graph_nodes.keys() == copy.graph_nodes.keys()
    assert original.artists.keys() == copy.artists.keys()


def do_test_copy_keeps_kwargs(Fig):
    da = data_array(ndim=1)
    original = Fig(
        LineView,
        Node(da),
        scale={'xx': 'log'},
        norm='log',
        grid=True,
        title='A nice title',
    )
    copy = copy_figure(original)
    assert copy.canvas.xscale == 'log'
    assert copy.canvas.yscale == 'log'
    assert copy.canvas.grid
    assert copy.canvas.title == 'A nice title'


def test_copy_static():
    do_test_copy(StaticFig)


def test_copy_static_keeps_kwargs():
    do_test_copy_keeps_kwargs(StaticFig)


@pytest.mark.usefixtures('_use_ipympl')
def test_copy_interactive():
    do_test_copy(InteractiveFig)


@pytest.mark.usefixtures('_use_ipympl')
def test_copy_interactive_keeps_kwargs():
    do_test_copy_keeps_kwargs(InteractiveFig)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from . import testing
from .factory import (
    data1d,
    data2d,
    data3d,
    data_array,
    dataset,
    dataset1d,
    dataset2d,
    histogram1d,
    histogram2d,
    histogram3d,
    random,
    scatter,
    variable,
)

__all__ = [
    "data1d",
    "data2d",
    "data3d",
    "data_array",
    "dataset",
    "dataset1d",
    "dataset2d",
    "histogram1d",
    "histogram2d",
    "histogram3d",
    "random",
    "scatter",
    "testing",
    "variable",
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import os
import tempfile

import pytest

from plopp import Node
from plopp.backends.pythreejs.figure import Figure
from plopp.data.testing import scatter
from plopp.graphics.scatter3dview import Scatter3dView


def test_log_norm_3d_toolbar_button():
    da = scatter()
    fig = Figure(Scatter3dView, Node(da), x='x', y='y', z='z', norm='log')
    assert fig.toolbar['lognorm'].value


def test_save_to_html():
    da = scatter()
    fig = Figure(Scatter3dView, Node(da), x='x', y='y', z='z', norm='log')
    with tempfile.TemporaryDirectory() as path:
        fname = os.path.join(path, 'plopp_fig3d.html')
        fig.save(filename=fname)
        assert os.path.isfile(fname)


def test_save_to_html_with_bad_extension_raises():
    da = scatter()
    fig = Figure(Scatter3dView, Node(da), x='x', y='y', z='z', norm='log')
    with pytest.raises(ValueError, match=r'File extension must be \.html'):
        fig.save(filename='plopp_fig3d.png')



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from typing import Literal

import scipp as sc

from .. import backends
from ..core.typing import FigureLike
from .graphicalview import GraphicalView


class LineView(GraphicalView):
    """
    View that makes a visual representation of one-dimensional data.
    It has a :class:`Canvas` and a specialized ``update`` function that generates
    :class:`Line` artists.

    Parameters
    ----------
    *nodes:
        The nodes that are attached to the view.
    norm:
        Control the scaling on the vertical axis.
    vmin:
        Lower bound for the vertical axis. If a number (without a unit) is supplied,
        it is assumed that the unit is the same as the current vertical axis unit.
    vmax:
        Upper bound for the vertical axis. If a number (without a unit) is supplied,
        it is assumed that the unit is the same as the current vertical axis unit.
    autoscale:
        The behavior of the axis limits. If ``auto``, the limits automatically
        adjusts every time the data changes. If ``grow``, the limits are allowed to
        grow with time but they do not shrink.
    scale:
        Control the scaling of the horizontal axis. For example, specify
        ``scale={'tof': 'log'}`` if you want log-scale for the ``tof`` dimension.
    errorbars:
        Hide errorbars if ``False``.
    mask_color:
        The color of the masks.
    aspect:
        Aspect ratio for the axes.
    grid:
        Show grid if ``True``.
    title:
        The figure title.
    figsize:
        The width and height of the figure, in inches.
    ax:
        If supplied, use these axes to create the figure. If none are supplied, the
        canvas will create its own axes.
    format:
        Format of the figure displayed in the Jupyter notebook. If ``None``, a SVG is
        created as long as the number of markers in the figure is not too large. If too
        many markers are drawn, a PNG image is created instead.
    legend:
        Show legend if ``True``. If ``legend`` is a tuple, it should contain the
        ``(x, y)`` coordinates of the legend's anchor point in axes coordinates.
    **kwargs:
        All other kwargs are forwarded to Matplotlib:

        - ``matplotlib.pyplot.plot`` for 1d data with a non bin-edge coordinate
        - ``matplotlib.pyplot.step`` for 1d data with a bin-edge coordinate
    """

    def __init__(
        self,
        *nodes,
        norm: Literal['linear', 'log'] = 'linear',
        vmin: sc.Variable | float | None = None,
        vmax: sc.Variable | float | None = None,
        autoscale: Literal['auto', 'grow'] = 'auto',
        scale: dict[str, str] | None = None,
        errorbars: bool = True,
        mask_color: str = 'black',
        aspect: Literal['auto', 'equal'] = 'auto',
        grid: bool = False,
        title: str | None = None,
        figsize: tuple[float, float] | None = None,
        format: Literal['svg', 'png'] | None = None,
        legend: bool | tuple[float, float] = True,
        **kwargs,
    ):
        super().__init__(*nodes)

        self._dims = {'x': None}
        self._ndim = 1
        self._scale = {} if scale is None else scale
        self._errorbars = errorbars
        self._mask_color = mask_color
        self._kwargs = kwargs
        self._repr_format = format
        self.canvas = backends.canvas2d(
            cbar=False,
            aspect=aspect,
            grid=grid,
            figsize=figsize,
            title=title,
            vmin=vmin,
            vmax=vmax,
            autoscale=autoscale,
            legend=legend,
            **kwargs,
        )
        self.canvas.yscale = norm
        self.render()
        self.canvas.finalize()

    def make_artist(self, new_values):
        return backends.line(
            canvas=self.canvas,
            data=new_values,
            number=len(self.artists),
            errorbars=self._errorbars,
            mask_color=self._mask_color,
            **self._kwargs,
        )


def linefigure(*args, **kwargs) -> FigureLike:
    """
    Create a figure to represent one-dimensional data from one or more graph node(s).

    .. versionadded:: 24.04.0

    Examples
    --------
    Create an input node and attach a ``linefigure`` as a view:

      >>> da = pp.data.data1d()
      >>> in_node = pp.Node(da)
      >>> fig = pp.linefigure(in_node)

    Visualize two data arrays on the same figure:

      >>> a = pp.data.data1d()
      >>> b = 3 * a
      >>> a_node = pp.Node(a)
      >>> b_node = pp.Node(b)
      >>> fig = pp.linefigure(a_node, b_node)

    With a customization argument to make the vertical scale logarithmic:

      >>> da = pp.data.data1d()
      >>> in_node = pp.Node(da)
      >>> fig = pp.linefigure(in_node, norm='log')
    """

    return backends.figure2d(LineView, *args, **kwargs)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)


import numpy as np
import scipp as sc

default_dim_list = ['x', 'y', 'z', 'time', 'temperature']


def variable(
    ndim: int = 1,
    variances: bool = False,
    dtype: str = 'float64',
    unit: str = 'm/s',
    dims: list[str] | None = None,
    dim_list: list[str] = default_dim_list,
) -> sc.Variable:
    """
    Generate a sample ``Variable`` containing data based on a sine function.

    Parameters
    ----------
    ndim:
        The number of dimensions.
    variances:
        Add variances to the output if ``True``.
    dtype:
        The output variable's data type.
    unit:
        The output variable's unit.
    dims:
        List of dimension labels. If ``None``, they will be auto-generated.
    dim_list:
        List of dimension labels to use if no ``dims`` are provided.
    """

    shapes = np.arange(50, 0, -10)[:ndim][::-1]
    if dims is None:
        dims = dim_list[:ndim][::-1]

    axes = [np.arange(shape, dtype=np.float64) for shape in shapes]
    pos = np.meshgrid(*axes, indexing='ij')
    radius = np.linalg.norm(np.array(pos), axis=0)
    a = np.sin(radius / 5.0)

    var = sc.array(dims=dims, values=a, unit=unit, dtype=dtype)
    if variances:
        var.variances = np.abs(np.random.normal(a * 0.1, 0.05))

    return var


def data_array(
    ndim: int = 1,
    variances: bool = False,
    binedges: bool = False,
    labels: bool = False,
    masks: bool = False,
    ragged: bool = False,
    dtype: str = 'float64',
    unit: str = 'm/s',
    dims: list[str] | None = None,
    dim_list: list[str] = default_dim_list,
) -> sc.DataArray:
    """
    Generate a sample ``DataArray`` containing data based on a sine function, with
    coordinates. Optionally add masks, labels, attributes.
    It is also possible to turn the coordinates into bin-edges, or make a ragged (2d)
    coordinate.

    Parameters
    ----------
    ndim:
        The number of dimensions.
    variances:
        Add variances to the output if ``True``.
    binedges:
        The output will have bin-edge coordinates instead of bin-centers if ``True``.
    labels:
        Add non-dimension coordinates if ``True``.
    masks:
        Add masks if ``True``.
    ragged:
        Make one of the coordinates two-dimensional.
    dtype:
        The output variable's data type.
    unit:
        The output variable's unit.
    dims:
        List of dimension labels. If ``None``, they will be auto-generated.
    dim_list:
        List of dimension labels to use if no ``dims`` are provided.
    """

    coord_units = dict(zip(dim_list, ['m', 'm', 'm', 's', 'K'], strict=True))

    data = variable(
        ndim=ndim,
        variances=variances,
        dims=dims,
        dtype=dtype,
        unit=unit,
        dim_list=dim_list,
    )

    coord_dict = {
        data.dims[i]: sc.arange(
            data.dims[i],
            data.shape[i] + binedges,
            unit=coord_units[data.dims[i]],
            dtype=np.float64,
        )
        for i in range(ndim)
    }
    mask_dict = {}

    if labels:
        coord_dict["lab"] = sc.linspace(
            data.dims[0], 101.0, 105.0, data.shape[0], unit='s'
        )
    if masks:
        mask_dict["mask"] = sc.array(
            dims=data.dims, values=np.where(data.values > 0, True, False)
        )

    if ragged:
        grid = []
        for i, dim in enumerate(data.dims):
            if binedges and (i < ndim - 1):
                grid.append(coord_dict[dim].values[:-1])
            else:
                grid.append(coord_dict[dim].values)
        mesh = np.meshgrid(*grid, indexing="ij")
        coord_dict[data.dims[-1]] = sc.array(
            dims=data.dims, values=mesh[-1] + np.indices(mesh[-1].shape)[0]
        )
    return sc.DataArray(data=data, coords=coord_dict, masks=mask_dict)


def dataset(entries: list[str] | None = None, **kwargs) -> sc.Dataset:
    """
    Generate a sample ``Dataset``. See :func:`data_array` for more options.

    Parameters
    ----------
    entries:
        A list of names for the elements of the dataset.
    **kwargs:
        All other keyword arguments are forwarded to :func:`data_array`.
    """
    if entries is None:
        entries = ['a', 'b']
    return sc.Dataset(
        {entry: (10.0 * np.random.random()) * data_array(**kwargs) for entry in entries}
    )


def scatter(npoints=500, scale=10.0, seed=1) -> sc.DataArray:
    """
    Generate some three-dimensional scatter data, based on a normal distribution.

    Parameters
    ----------
    npoints:
        The number of points to generate.
    scale:
        Standard deviation (spread or 'width') of the distribution.
    seed:
        The seed for the random number generator.
    """
    rng = np.random.default_rng(seed)
    position = scale * rng.standard_normal(size=[npoints, 3])
    values = np.linalg.norm(position, axis=1)
    vec = sc.vectors(dims=['row'], unit='m', values=position)
    return sc.DataArray(
        data=sc.array(dims=['row'], values=values, unit='K'),
        coords={
            'position': vec,
            'x': vec.fields.x,
            'y': vec.fields.y,
            'z': vec.fields.z,
        },
    )


def random(shape, dtype='float64', unit='', dims=None, seed=None) -> sc.DataArray:
    """
    Generate a data array containing random data values.

    Parameters
    ----------
    shape:
        The shape of the output.
    dtype:
        The output data array's data type.
    unit:
        The output data array's unit.
    dims:
        List of dimension labels. If ``None``, they will be auto-generated.
    seed:
        The seed for the random number generator.
    """
    rng = np.random.default_rng(seed)
    if dims is None:
        dims = default_dim_list[: len(shape)][::-1]
    return sc.DataArray(
        data=sc.array(
            dims=dims,
            values=rng.random(shape),
            unit=unit,
            dtype=dtype,
        ),
        coords={
            dim: sc.arange(dim, shape[i], unit='m', dtype='float64')
            for i, dim in enumerate(dims)
        },
    )


def data1d(**kwargs):
    return data_array(ndim=1, **kwargs)


def data2d(**kwargs):
    return data_array(ndim=2, **kwargs)


def data3d(**kwargs):
    return data_array(ndim=3, **kwargs)


def histogram1d(**kwargs):
    return data_array(ndim=1, binedges=True, **kwargs)


def histogram2d(**kwargs):
    return data_array(ndim=2, binedges=True, **kwargs)


def histogram3d(**kwargs):
    return data_array(ndim=3, binedges=True, **kwargs)


def dataset1d(**kwargs):
    return dataset(ndim=1, **kwargs)


def dataset2d(**kwargs):
    return dataset(ndim=2, **kwargs)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from typing import Literal

import numpy as np
import scipp as sc

from .utils import merge_masks


def find_limits(
    x: sc.DataArray,
    scale: Literal['linear', 'log'] = 'linear',
    pad: bool = False,
) -> tuple[sc.Variable, sc.Variable]:
    """
    Find sensible limits, depending on linear or log scale.
    If there are no finite values in the array, raise an error.
    If there are no positive values in the array, and the scale is log, fall back to
    some sensible default values.
    """
    is_datetime = x.dtype == sc.DType.datetime64
    # Computing limits for string arrays is not supported, so we convert them to
    # dummy numerical arrays.
    if x.dtype == sc.DType.string:
        x = x.copy(deep=False)
        x.data = sc.arange(x.dim, float(len(x)), unit=x.unit)
    if x.masks:
        x = sc.where(
            merge_masks(x.masks),
            sc.scalar(np.nan, unit=x.unit),
            x.data.to(dtype='float64'),
        )
    v = x.values
    finite_inds = np.isfinite(v)
    if np.sum(finite_inds) == 0:
        raise ValueError("No finite values were found in array. Cannot compute limits.")
    finite_vals = v[finite_inds]
    finite_max = None
    if scale == "log":
        if is_datetime:
            finite_min = np.amin(finite_vals)
        else:
            positives = finite_vals > 0
            if np.sum(positives) == 0:
                finite_min = 0.1
                finite_max = 1.0
            else:
                initial = (np.finfo if v.dtype.kind == 'f' else np.iinfo)(v.dtype).max
                finite_min = np.amin(
                    finite_vals, initial=initial, where=finite_vals > 0
                )
    else:
        finite_min = np.amin(finite_vals)
    if finite_max is None:
        finite_max = np.amax(finite_vals)
    if pad:
        delta = 0.05
        if (scale == 'log') and (not is_datetime):
            p = (finite_max / finite_min) ** delta
            finite_min /= p
            finite_max *= p
        if scale == 'linear':
            p = (finite_max - finite_min) * delta
            finite_min -= p
            finite_max += p
    return (sc.scalar(finite_min, unit=x.unit), sc.scalar(finite_max, unit=x.unit))


def fix_empty_range(
    lims: tuple[sc.Variable, sc.Variable],
) -> tuple[sc.Variable, sc.Variable]:
    """
    Range correction in case xmin == xmax
    """
    if lims[0].value != lims[1].value:
        return lims
    if lims[0].value == 0.0:
        dx = sc.scalar(0.5, unit=lims[0].unit)
    else:
        # We decompose value and unit to avoid operation exceptions when unit is None.
        dx = sc.scalar(0.5 * abs(lims[0].value), unit=lims[0].unit)
    return (lims[0] - dx, lims[1] + dx)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import plopp as pp


def test_plot_data_array_1d():
    pp.plot(pp.data.data_array(ndim=1))


def test_plot_data_array_2d():
    pp.plot(pp.data.data_array(ndim=2))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

DEFAULTS = {'2d': 'matplotlib', '3d': 'pythreejs'}


class BackendManager:
    def __init__(self):
        self._mapping = DEFAULTS.copy()
        self._backends = {}
        self._sync()

    def __getitem__(self, key):
        return self._mapping[key]

    def __setitem__(self, key, value):
        self._mapping[key] = value
        self._sync()

    def keys(self):
        return self._mapping.keys()

    def values(self):
        return self._mapping.values()

    def items(self):
        return self._mapping.items()

    def reset(self):
        self._mapping = DEFAULTS.copy()
        self._sync()

    def _sync(self):
        if self._mapping['2d'] == 'matplotlib':
            from .matplotlib import MatplotlibBackend

            self._backends['2d'] = MatplotlibBackend()
        elif self._mapping['2d'] == 'plotly':
            from .plotly import PlotlyBackend

            self._backends['2d'] = PlotlyBackend()
        else:
            raise ValueError(f'Unsupported 2d backend \'{self._mapping["2d"]}\'.')

        if self._mapping['3d'] == 'pythreejs':
            from .pythreejs import PythreejsBackend

            self._backends['3d'] = PythreejsBackend()
        else:
            raise ValueError(f'Unsupported 3d backend \'{self._mapping["3d"]}\'.')

    def is_interactive(self):
        return self._backends['2d'].is_interactive()

    def canvas2d(self, *args, **kwargs):
        try:
            _canvas2d = self._backends['2d'].canvas2d
        except AttributeError:
            raise ValueError(
                f'Unsupported backend \'{self["2d"]}\' for canvas2d.'
            ) from None
        return _canvas2d(*args, **kwargs)

    def canvas3d(self, *args, **kwargs):
        try:
            _canvas3d = self._backends['3d'].canvas3d
        except AttributeError:
            raise ValueError(
                f'Unsupported backend \'{self["3d"]}\' for canvas3d.'
            ) from None
        return _canvas3d(*args, **kwargs)

    def line(self, *args, **kwargs):
        try:
            _line = self._backends['2d'].line
        except AttributeError:
            raise ValueError(
                f'Unsupported backend \'{self["2d"]}\' for line (1D).'
            ) from None
        return _line(*args, **kwargs)

    def image(self, *args, **kwargs):
        try:
            _image = self._backends['2d'].image
        except AttributeError:
            raise ValueError(
                f'Unsupported backend \'{self["2d"]}\' for image (2D).'
            ) from None
        return _image(*args, **kwargs)

    def scatter(self, *args, **kwargs):
        try:
            _scatter = self._backends['2d'].scatter
        except AttributeError:
            raise ValueError(
                f'Unsupported backend \'{self["2d"]}\' for scatter.'
            ) from None
        return _scatter(*args, **kwargs)

    def point_cloud(self, *args, **kwargs):
        try:
            _point_cloud = self._backends['3d'].point_cloud
        except AttributeError:
            raise ValueError(
                f'Unsupported backend \'{self["3d"]}\' for point_cloud (3D).'
            ) from None
        return _point_cloud(*args, **kwargs)

    def figure1d(self, *args, **kwargs):
        try:
            _figure1d = self._backends['2d'].figure1d
        except AttributeError:
            raise ValueError(
                f'Unsupported backend \'{self["2d"]}\' for figure1d.'
            ) from None
        return _figure1d(*args, **kwargs)

    def figure2d(self, *args, **kwargs):
        try:
            _figure2d = self._backends['2d'].figure2d
        except AttributeError:
            raise ValueError(
                f'Unsupported backend \'{self["2d"]}\' for figure2d.'
            ) from None
        return _figure2d(*args, **kwargs)

    def figure3d(self, *args, **kwargs):
        try:
            _figure3d = self._backends['3d'].figure3d
        except AttributeError:
            raise ValueError(
                f'Unsupported backend \'{self["3d"]}\' for figure3d.'
            ) from None
        return _figure3d(*args, **kwargs)

    def tiled(self, *args, **kwargs):
        try:
            _tiled = self._backends['2d'].tiled
        except AttributeError:
            raise ValueError(
                f'Unsupported backend \'{self["2d"]}\' for tiled.'
            ) from None
        return _tiled(*args, **kwargs)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from collections.abc import Iterable
from copy import copy
from functools import reduce
from typing import Any, Literal

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import scipp as sc
from matplotlib.colorbar import ColorbarBase
from matplotlib.colors import Colormap, LinearSegmentedColormap, LogNorm, Normalize

from ..backends.matplotlib.utils import fig_to_bytes
from ..core.limits import find_limits, fix_empty_range
from ..core.utils import maybe_variable_to_number, merge_masks


def _get_cmap(name: str, nan_color: str | None = None) -> Colormap:
    """
    Get a colormap object from a colormap name.

    Parameters
    ----------
    name:
        Name of the colormap. If the name is just a single html color, this will
        create a colormap with that single color.
    nan_color:
        The color to use for NAN values.
    """

    try:
        if hasattr(mpl, 'colormaps'):
            cmap = copy(mpl.colormaps[name])
        else:
            cmap = mpl.cm.get_cmap(name)
    except (KeyError, ValueError):
        cmap = LinearSegmentedColormap.from_list('tmp', [name, name])
    # TODO: we need to set under and over values for the cmap with
    # `cmap.set_under` and `cmap.set_over`. Ideally these should come from a config?
    if nan_color is not None:
        cmap.set_bad(color=nan_color)
    return cmap


def _get_normalizer(norm: str) -> Normalize:
    """
    Get an appropriate normalizer depending on the scaling.
    """
    return Normalize(vmin=0, vmax=1) if norm == 'linear' else LogNorm(vmin=1, vmax=2)


class ColorMapper:
    """
    A class that handles conversion between data values and RGBA colors.
    It controls the normalization (linear or log), as well as the min and max limits
    for the color range.

    Parameters
    ----------
    cax:
        The axes to use for the colorbar. If none are supplied, the ColorMapper will
        create its own axes.
    cbar:
        Create a colorbar if ``True``. If ``False``, no colorbar is made even if ``cax``
        is defined.
    cmap:
        The name of the colormap for the data
        (see https://matplotlib.org/stable/tutorials/colors/colormaps.html).
        In addition to the Matplotlib docs, if the name is just a single html color,
        a colormap with that single color will be used.
    mask_cmap:
        The name of the colormap for masked data.
    norm:
        The colorscale normalization.
    autoscale:
        The behavior of the color range limits. If ``auto``, the limits automatically
        adjusts every time the data changes. If ``grow``, the limits are allowed to
        grow with time but they do not shrink.
    vmin:
        The minimum value for the colorscale range. If a number (without a unit) is
        supplied, it is assumed that the unit is the same as the data unit.
    vmax:
        The maximum value for the colorscale range. If a number (without a unit) is
        supplied, it is assumed that the unit is the same as the data unit.
    nan_color:
        The color used for representing NAN values.
    figsize:
        The size of the figure next to which the colorbar will be displayed.
    """

    def __init__(
        self,
        canvas: Any | None = None,
        cbar: bool = True,
        cmap: str = 'viridis',
        mask_cmap: str = 'gray',
        norm: Literal['linear', 'log'] = 'linear',
        autoscale: Literal['auto', 'grow'] = 'auto',
        vmin: sc.Variable | float | None = None,
        vmax: sc.Variable | float | None = None,
        nan_color: str | None = None,
        figsize: tuple[float, float] | None = None,
    ):
        self.cax = canvas.cax if canvas is not None else None
        self.cmap = _get_cmap(cmap, nan_color=nan_color)
        self.mask_cmap = _get_cmap(mask_cmap, nan_color=nan_color)
        self.user_vmin = vmin
        self.user_vmax = vmax
        self.vmin = np.inf
        self.vmax = -np.inf
        self.norm = norm
        self._autoscale = autoscale

        # Note that we need to set vmin/vmax for the LogNorm, if not an error is
        # raised when making the colorbar before any call to update is made.
        self.normalizer = _get_normalizer(self.norm)
        self.colorbar = None
        self.unit = None

        self.name = None
        self.changed = False
        self.artists = {}
        self.widget = None

        if cbar:
            if self.cax is None:
                dpi = 100
                height_inches = (figsize[1] / dpi) if figsize is not None else 6
                fig = plt.Figure(figsize=(height_inches * 0.2, height_inches))
                self.cax = fig.add_axes([0.05, 0.02, 0.2, 0.98])
            self.colorbar = ColorbarBase(self.cax, cmap=self.cmap, norm=self.normalizer)
            self.cax.yaxis.set_label_coords(-0.9, 0.5)

    def __setitem__(self, key: str, artist: Any):
        self.artists[key] = artist

    def __getitem__(self, key: str) -> Any:
        return self.artists[key]

    def to_widget(self):
        """
        Convert the colorbar into a widget for use with other ``ipywidgets``.
        """
        import ipywidgets as ipw

        self.widget = ipw.HTML()
        self._update_colorbar_widget()
        return self.widget

    def _update_colorbar_widget(self):
        """
        Upon an updated colorscale range, we need to update the image inside the widget.
        """
        if self.widget is not None:
            self.widget.value = fig_to_bytes(self.cax.get_figure(), form='svg').decode()

    def rgba(self, data: sc.DataArray) -> np.ndarray:
        """
        Return rgba values given a data array.

        Parameters
        ----------
        data:
            The data array to be converted to rgba colors, taking masks into account.
        """
        colors = self.cmap(self.normalizer(data.values))
        if data.masks:
            one_mask = merge_masks(data.masks).values
            colors[one_mask] = self.mask_cmap(self.normalizer(data.values[one_mask]))
        return colors

    def autoscale(self):
        """
        Re-compute the global min and max range of values by iterating over all the
        artists. If autoscale is set to ``'auto'``, the limits adjust to he current
        range. If it is set to ``'grow'``, limits can grow but not shrink.
        """
        limits = [
            fix_empty_range(find_limits(artist._data, scale=self.norm))
            for artist in self.artists.values()
        ]
        vmin = reduce(min, [v[0] for v in limits])
        vmax = reduce(max, [v[1] for v in limits])
        if self.user_vmin is not None:
            self.vmin = self.user_vmin
        elif (vmin.value < self.vmin) or (self._autoscale == 'auto'):
            self.vmin = vmin.value
        if self.user_vmax is not None:
            self.vmax = self.user_vmax
        elif (vmax.value > self.vmax) or (self._autoscale == 'auto'):
            self.vmax = vmax.value

        if self.vmin >= self.vmax:
            if self.user_vmax is not None:
                self.vmax = self.user_vmax
                self.vmin = self.user_vmax - abs(self.user_vmax) * 0.1
            else:
                self.vmin = self.user_vmin
                self.vmax = self.user_vmin + abs(self.user_vmin) * 0.1

    def _set_artists_colors(self, keys: Iterable):
        """
        Update the colors of all the artists apart from the one that triggered the
        update, as those get updated by the figure.

        Parameters
        ----------
        keys:
            List of artists to update.
        """
        for k in keys:
            self.artists[k].set_colors(self.rgba(self.artists[k].data))

    def _set_normalizer_limits(self):
        """
        Synchronize the underlying normalizer limits to the current state.
        """
        # Note that the order matters here, as for a normalizer vmin cannot be set above
        # the current vmax.
        if self.vmin >= self.normalizer.vmax:
            self.normalizer.vmax = self.vmax
            self.normalizer.vmin = self.vmin
        else:
            self.normalizer.vmin = self.vmin
            self.normalizer.vmax = self.vmax

    def update(self, *args, **kwargs):
        """
        Update the colorscale bounds taking into account new values,
        by either supplying a dictionary of new data or by keyword arguments.
        We also update the colorbar widget if it exists.
        """
        new = dict(*args, **kwargs)
        for key, data in new.items():
            if self.name is None:
                self.name = data.name
                # If name is None, this is the first time update is called
                if self.user_vmin is not None:
                    self.user_vmin = maybe_variable_to_number(
                        self.user_vmin, unit=self.unit
                    )
                if self.user_vmax is not None:
                    self.user_vmax = maybe_variable_to_number(
                        self.user_vmax, unit=self.unit
                    )
            elif data.name != self.name:
                self.name = ''
            if self.cax is not None:
                text = self.name
                if self.unit is not None:
                    text += f'{" " if self.name else ""}[{self.unit}]'
                self.cax.set_ylabel(text)
            old_bounds = np.array([self.vmin, self.vmax])
            self.autoscale()
            self._set_normalizer_limits()

            if not np.allclose(old_bounds, np.array([self.vmin, self.vmax])):
                self._update_colorbar_widget()
                keys = self.artists.keys()
            else:
                keys = [key]
            self._set_artists_colors(keys)

    def toggle_norm(self):
        """
        Toggle the norm flag, between `linear` and `log`.
        """
        self.norm = "log" if self.norm == 'linear' else 'linear'
        self.normalizer = _get_normalizer(self.norm)
        self.vmin = np.inf
        self.vmax = -np.inf
        self.autoscale()
        self._set_normalizer_limits()
        self._set_artists_colors(self.artists.keys())
        if self.colorbar is not None:
            self.colorbar.mappable.norm = self.normalizer
            self._update_colorbar_widget()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from __future__ import annotations

from typing import Any

import numpy as np
from matplotlib import gridspec

from ...core.typing import FigureLike
from .static import get_repr_maker
from .utils import copy_figure, is_interactive_backend, make_figure


class Tiled:
    """
    A tiled figure.
    This is based on Matpotlib's GridSpec.

    .. versionadded:: 23.08.0

    Parameters
    ----------
    nrows:
        Number of rows.
    ncols:
        Number of columns.
    figsize:
        Figure size (width, height) in inches.
    **kwargs:
        Additional arguments passed to :class:`matplotlib.gridspec.GridSpec`.

    Examples
    --------
    Create a tiled figure with two plots stacked vertically:

      >>> da1 = pp.data.data1d()
      >>> da2 = pp.data.data2d()
      >>> tiled = pp.tiled(2, 1)
      >>> tiled[0] = da1.plot()
      >>> tiled[1] = da2.plot()

    Create a tiled 2x2 figure:

      >>> da1 = pp.data.data1d()
      >>> da2 = pp.data.data2d()
      >>> da3 = pp.data.data2d()
      >>> da4 = pp.data.data1d()
      >>> tiled = pp.tiled(2, 2)
      >>> tiled[0, 0] = da1.plot()
      >>> tiled[0, 1] = da2.plot()
      >>> tiled[1, 0] = da3.plot()
      >>> tiled[1, 1] = da4.plot()

    Create a tiled figure with two figures side by side and the first is twice as wide:

      >>> da1 = pp.data.data1d()
      >>> da2 = pp.data.data2d()
      >>> tiled = pp.tiled(1, 3)
      >>> tiled[0, :2] = da1.plot()
      >>> tiled[0, 2] = da2.plot()

    """

    def __init__(
        self,
        nrows: int,
        ncols: int,
        figsize: tuple[float, float] | None = None,
        **kwargs: Any,
    ) -> None:
        self.nrows = nrows
        self.ncols = ncols
        self.fig = make_figure(
            figsize=(
                (min(6.0 * ncols, 15.0), min(4.0 * nrows, 15.0))
                if figsize is None
                else figsize
            ),
            layout='constrained',
        )

        self.gs = gridspec.GridSpec(nrows, ncols, figure=self.fig, **kwargs)
        self.axes = []
        self.views = np.full((nrows, ncols), None)
        self._history = []

    def __setitem__(
        self,
        inds: int | slice | tuple[int, int] | tuple[slice, slice],
        view: FigureLike,
    ) -> None:
        new_view = copy_figure(view, ax=self.fig.add_subplot(self.gs[inds]))
        self.views[inds] = new_view
        self._history.append((inds, new_view))

    def __getitem__(
        self, inds: int | slice | tuple[int, int] | tuple[slice, slice]
    ) -> FigureLike:
        return self.views[inds]

    def _repr_mimebundle_(self, include=None, exclude=None) -> dict:
        """
        Mimebundle display representation for jupyter notebooks.
        """
        if is_interactive_backend():
            return self.fig.canvas._repr_mimebundle_(include=include, exclude=exclude)
        else:
            out = {'text/plain': f'TiledFigure(nrows={self.nrows}, ncols={self.ncols})'}
            npoints = sum(
                len(line.get_xdata()) for ax in self.axes for line in ax.lines
            )
            out.update(get_repr_maker(npoints=npoints)(self.fig))
            return out

    def save(self, filename: str, **kwargs: Any) -> None:
        """
        Save the figure to file.
        The default directory for writing the file is the same as the
        directory where the script or notebook is running.

        Parameters
        ----------
        filename:
            Name of the output file. Possible file extensions are ``.jpg``, ``.png``,
            ``.svg``, and ``.pdf``.
        """
        self.fig.savefig(filename, **{**{'bbox_inches': 'tight'}, **kwargs})

    def show(self) -> None:
        """
        Make a call to Matplotlib's underlying ``show`` function.
        """
        self.fig.show()

    def __add__(self, other: Tiled) -> Tiled:
        if not isinstance(other, self.__class__):
            t = Tiled(1, 1)
            t[0, 0] = other
            other = t

        out = Tiled(nrows=max(self.nrows, other.nrows), ncols=self.ncols + other.ncols)
        for inds, view in self._history:
            out[inds] = view
        for inds, view in other._history:
            out[inds[0], inds[1] + self.ncols] = view
        return out

    def __truediv__(self, other: Tiled) -> Tiled:
        if not isinstance(other, self.__class__):
            t = Tiled(1, 1)
            t[0, 0] = other
            other = t

        out = Tiled(nrows=self.nrows + other.nrows, ncols=max(self.ncols, other.ncols))
        for inds, view in self._history:
            out[inds] = view
        for inds, view in other._history:
            out[inds[0] + self.nrows, inds[1]] = view
        return out


def hstack(left: Tiled | FigureLike, right: Tiled | FigureLike) -> Tiled:
    """
    Display two views side by side.

    Parameters
    ----------
    left:
        The view to display on the left.
    right:
        The view to display on the right.
    """
    left_tiled = isinstance(left, Tiled)
    right_tiled = isinstance(right, Tiled)
    if (not left_tiled) and (not right_tiled):
        out = Tiled(1, 2)
        out[0, 0] = left
        out[0, 1] = right
        return out
    elif left_tiled:
        t = Tiled(1, 1)
        t[0, 0] = right
        return left + t
    else:
        t = Tiled(1, 1)
        t[0, 0] = left
        return t + right


def vstack(top: Tiled | FigureLike, bottom: Tiled | FigureLike) -> Tiled:
    """
    Display two views on top of each other.

    Parameters
    ----------
    top:
        The view to display on the top.
    bottom:
        The view to display on the bottom.
    """
    top_tiled = isinstance(top, Tiled)
    bottom_tiled = isinstance(bottom, Tiled)
    if (not top_tiled) and (not bottom_tiled):
        out = Tiled(2, 1)
        out[0, 0] = top
        out[1, 0] = bottom
        return out
    elif top_tiled:
        t = Tiled(1, 1)
        t[0, 0] = bottom
        return top / t
    else:
        t = Tiled(1, 1)
        t[0, 0] = top
        return t / bottom



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import uuid
from functools import partial
from typing import Literal

import scipp as sc

from ..core.typing import FigureLike, PlottableMulti
from ..graphics import Camera
from .common import check_not_binned, from_compatible_lib, input_to_nodes


def _preprocess_scatter(
    obj: PlottableMulti,
    x: str,
    y: str,
    z: str,
    pos: str | None,
    name: str | None = None,
) -> sc.DataArray:
    da = from_compatible_lib(obj)
    check_not_binned(da)

    if pos is not None:
        coords = {k: getattr(da.coords[pos].fields, k) for k in (x, y, z)}
    else:
        coords = {k: da.coords[k] for k in (x, y, z)}

    out = sc.DataArray(data=da.data, masks=da.masks, coords=coords)
    if out.ndim != 1:
        out = out.flatten(to=uuid.uuid4().hex)
    if name is not None:
        out.name = name
    return out


def scatter3d(
    obj: PlottableMulti,
    *,
    x: str = 'x',
    y: str = 'y',
    z: str = 'z',
    pos: str | None = None,
    figsize: tuple[int, int] = (600, 400),
    norm: Literal['linear', 'log'] = 'linear',
    title: str | None = None,
    vmin: sc.Variable | float = None,
    vmax: sc.Variable | float = None,
    cmap: str = 'viridis',
    camera: Camera | None = None,
    **kwargs,
) -> FigureLike:
    """Make a three-dimensional scatter plot.

    To specify the positions of the scatter points, you can use:

    - a single coordinate inside the supplied data array that has dtype ``vector3``
      (use the ``pos`` parameter to specify the name of the coordinate).
    - three coordinates from the data array, whose names are specified using the
      ``x``, ``y``, and ``z`` arguments.

    Note that if ``pos`` is used, ``x``, ``y``, and ``z`` must all be ``None``.

    Parameters
    ----------
    obj:
        The data array containing the data and the coordinates.
    x:
        The name of the coordinate that is to be used for the X positions.
    y:
        The name of the coordinate that is to be used for the Y positions.
    z:
        The name of the coordinate that is to be used for the Z positions.
    pos:
        The name of the vector coordinate that is to be used for the positions.
    norm:
        Set to ``'log'`` for a logarithmic colorscale.
    figsize:
        The size of the 3d rendering area, in pixels: ``(width, height)``.
    title:
        The figure title.
    vmin:
        Lower bound for the colorscale.
    vmax:
        Upper bound for the colorscale.
    cmap:
        The name of the colormap.
    camera:
        Initial camera configuration (position, target).

    Returns
    -------
    :
        A three-dimensional interactive scatter plot.
    """
    from ..graphics import scatter3dfigure
    from ..widgets import ClippingPlanes, ToggleTool

    if 'ax' in kwargs:
        raise ValueError(
            'Keyword "ax" detected. Embedding 3D scatter plots inside Matplotlib axes '
            'is not supported. See '
            'https://scipp.github.io/plopp/customization/subplots.html#FAQ:-subplots-with-3D-scatter-plots'
        )

    nodes = input_to_nodes(
        obj, processor=partial(_preprocess_scatter, x=x, y=y, z=z, pos=pos)
    )

    fig = scatter3dfigure(
        *nodes,
        x=x,
        y=y,
        z=z,
        figsize=figsize,
        norm=norm,
        title=title,
        vmin=vmin,
        vmax=vmax,
        cmap=cmap,
        camera=camera,
        **kwargs,
    )
    clip_planes = ClippingPlanes(fig)
    fig.toolbar['cut3d'] = ToggleTool(
        callback=clip_planes.toggle_visibility,
        icon='layer-group',
        tooltip='Hide/show spatial cutting tool',
    )
    fig.bottom_bar.add(clip_planes)
    return fig



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import os

from ipywidgets import HBox, VBox

from ...graphics import BaseFig
from ...widgets import HBar, VBar, make_toolbar_canvas3d


class Figure(BaseFig, VBox):
    """
    Create a figure to represent three-dimensional data.
    """

    def __init__(self, View, *args, **kwargs):
        self._view = View(*args, **kwargs)
        self.toolbar = make_toolbar_canvas3d(
            canvas=self._view.canvas, colormapper=self._view.colormapper
        )
        self.left_bar = VBar([self.toolbar])
        self.right_bar = VBar([self._view.colormapper.to_widget()])
        self.bottom_bar = HBar()
        self.top_bar = HBar([self._view.canvas._title])

        super().__init__(
            [
                self.top_bar,
                HBox([self.left_bar, self._view.canvas.to_widget(), self.right_bar]),
                self.bottom_bar,
            ]
        )

    def save(self, filename):
        """
        Save the figure to a standalone HTML file.
        The default directory for writing the file is the same as the
        directory where the script or notebook is running.

        Parameters
        ----------
        filename:
            Name of the output HTML file.
        """
        ext = os.path.splitext(filename)[1]
        if ext.lower() != '.html':
            raise ValueError('File extension must be .html for saving 3d figures.')
        from ipywidgets.embed import dependency_state, embed_minimal_html

        out = HBox([self._view.canvas.to_widget(), self.right_bar])
        # Garbage collection for embedded html output:
        # https://github.com/jupyter-widgets/pythreejs/issues/217
        state = dependency_state(out)
        # convert and write to file
        embed_minimal_html(
            filename,
            out,
            title=self._view.canvas.title if self._view.canvas.title else 'figure3d',
            state=state,
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from .box import Box, HBar, VBar
from .checkboxes import Checkboxes
from .clip3d import Clip3dTool, ClippingPlanes
from .cut3d import Cut3dTool, TriCutTool
from .drawing import DrawingTool, PointsTool
from .linesave import LineSaveTool
from .slice import RangeSliceWidget, SliceWidget, slice_dims
from .toolbar import Toolbar, make_toolbar_canvas2d, make_toolbar_canvas3d
from .tools import ButtonTool, ColorTool, ToggleTool

__all__ = [
    "Box",
    "ButtonTool",
    "Checkboxes",
    "Clip3dTool",
    "ClippingPlanes",
    "ColorTool",
    "Cut3dTool",
    "DrawingTool",
    "HBar",
    "LineSaveTool",
    "PointsTool",
    "RangeSliceWidget",
    "SliceWidget",
    "ToggleTool",
    "Toolbar",
    "TriCutTool",
    "VBar",
    "make_toolbar_canvas2d",
    "make_toolbar_canvas3d",
    "slice_dims",
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import ipywidgets as ipw
import scipp as sc

from plopp import Node, imagefigure, linefigure, node, widget_node
from plopp.data.testing import data_array, dataset
from plopp.widgets import Box, Checkboxes, SliceWidget, slice_dims


@node
def hide_masks(data_array, masks):
    out = data_array.copy(deep=False)
    for name, value in masks.items():
        if name in out.masks and (not value):
            del out.masks[name]
    return out


def test_single_1d_line():
    da = data_array(ndim=1)
    n = Node(da)
    _ = linefigure(n)


def test_two_1d_lines():
    ds = dataset(ndim=1)
    a = Node(ds['a'])
    b = Node(ds['b'])
    _ = linefigure(a, b)


def test_difference_of_two_1d_lines():
    ds = dataset(ndim=1)
    a = Node(ds['a'])
    b = Node(ds['b'])

    @node
    def diff(x, y):
        return x - y

    c = diff(a, b)
    _ = linefigure(a, b, c)


def test_2d_image():
    da = data_array(ndim=2)
    a = Node(da)
    _ = imagefigure(a)


def test_2d_image_smoothing_slider():
    da = data_array(ndim=2)
    a = Node(da)

    sl = ipw.IntSlider(min=1, max=10)
    sigma_node = widget_node(sl)

    from scipp.scipy.ndimage import gaussian_filter

    smooth_node = Node(gaussian_filter, a, sigma=sigma_node)

    fig = imagefigure(smooth_node)
    Box([fig.to_widget(), sl])
    sl.value = 5


def test_2d_image_with_masks():
    da = data_array(ndim=2)
    da.masks['m1'] = da.data < sc.scalar(0.0, unit='m/s')
    da.masks['m2'] = da.coords['xx'] > sc.scalar(30.0, unit='m')

    a = Node(da)

    widget = Checkboxes(da.masks.keys())
    w = widget_node(widget)

    masks_node = hide_masks(a, w)
    fig = imagefigure(masks_node)
    Box([fig.to_widget(), widget])
    widget.toggle_all_button.value = False


def test_two_1d_lines_with_masks():
    ds = dataset()
    ds['a'].masks['m1'] = ds['a'].coords['xx'] > sc.scalar(40.0, unit='m')
    ds['a'].masks['m2'] = ds['a'].data < ds['b'].data
    ds['b'].masks['m1'] = ds['b'].coords['xx'] < sc.scalar(5.0, unit='m')

    a = Node(ds['a'])
    b = Node(ds['b'])

    widget = Checkboxes(list(ds['a'].masks.keys()) + list(ds['b'].masks.keys()))
    w = widget_node(widget)

    node_masks_a = hide_masks(a, w)
    node_masks_b = hide_masks(b, w)
    fig = linefigure(node_masks_a, node_masks_b)
    Box([fig.to_widget(), widget])
    widget.toggle_all_button.value = False


def test_node_sum_data_along_y():
    da = data_array(ndim=2, binedges=True)
    a = Node(da)
    s = Node(sc.sum, a, dim='yy')

    fig1 = imagefigure(a)
    fig2 = linefigure(s)
    Box([[fig1.to_widget(), fig2.to_widget()]])


def test_slice_3d_cube():
    da = data_array(ndim=3)
    a = Node(da)
    sl = SliceWidget(da, dims=['zz'])
    w = widget_node(sl)

    slice_node = slice_dims(a, w)

    fig = imagefigure(slice_node)
    Box([fig.to_widget(), sl])
    sl.controls["zz"]["slider"].value = 10


def test_3d_image_slicer_with_connected_side_histograms():
    da = data_array(ndim=3)
    a = Node(da)
    sl = SliceWidget(da, dims=['zz'])
    w = widget_node(sl)

    sliced = slice_dims(a, w)
    fig = imagefigure(sliced)

    histx = Node(sc.sum, sliced, dim='xx')
    histy = Node(sc.sum, sliced, dim='yy')

    fx = linefigure(histx)
    fy = linefigure(histy)
    Box([[fx.to_widget(), fy.to_widget()], fig.to_widget(), sl])
    sl.controls["zz"]["slider"].value = 10



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import numpy as np
import scipp as sc

from plopp.backends.matplotlib.canvas import Canvas
from plopp.backends.matplotlib.line import Line
from plopp.data.testing import data_array


def test_line_creation():
    da = data_array(ndim=1, unit='K')
    line = Line(canvas=Canvas(), data=da)
    assert line._unit == 'K'
    assert line._dim == 'xx'
    assert len(line._line.get_xdata()) == da.sizes['xx']
    assert np.allclose(line._line.get_xdata(), da.coords['xx'].values)
    assert np.allclose(line._line.get_ydata(), da.values)
    assert line._error is None
    assert not line._mask.get_visible()


def test_line_creation_bin_edges():
    da = data_array(ndim=1, binedges=True)
    line = Line(canvas=Canvas(), data=da)
    assert len(line._line.get_xdata()) == da.sizes['xx'] + 1


def test_line_with_errorbars():
    da = data_array(ndim=1, variances=True)
    line = Line(canvas=Canvas(), data=da)
    assert line._error.has_yerr
    coll = line._error.get_children()[0]
    x = np.array(coll.get_segments())[:, 0, 0]
    y1 = np.array(coll.get_segments())[:, 0, 1]
    y2 = np.array(coll.get_segments())[:, 1, 1]
    assert np.allclose(x, da.coords['xx'].values)
    assert np.allclose(y1, (da.data - sc.stddevs(da.data)).values)
    assert np.allclose(y2, (da.data + sc.stddevs(da.data)).values)


def test_line_with_bin_edges_and_errorbars():
    da = data_array(ndim=1, binedges=True, variances=True)
    line = Line(canvas=Canvas(), data=da)
    coll = line._error.get_children()[0]
    x = np.array(coll.get_segments())[:, 0, 0]
    assert np.allclose(x, sc.midpoints(da.coords['xx']).values)


def test_line_hide_errorbars():
    da = data_array(ndim=1, variances=True)
    line = Line(canvas=Canvas(), data=da, errorbars=False)
    assert line._error is None


def test_line_with_mask():
    da = data_array(ndim=1, masks=True)
    line = Line(canvas=Canvas(), data=da)
    assert line._mask.get_visible()


def test_line_with_mask_and_binedges():
    da = data_array(ndim=1, binedges=True, masks=True)
    line = Line(canvas=Canvas(), data=da)
    assert line._mask.get_visible()


def test_line_with_two_masks():
    da = data_array(ndim=1, masks=True)
    da.masks['two'] = da.coords['xx'] > sc.scalar(25, unit='m')
    line = Line(canvas=Canvas(), data=da)
    expected = da.data[da.masks['mask'] | da.masks['two']].values
    y = line._mask.get_ydata()
    assert np.allclose(y[~np.isnan(y)], expected)


def test_line_update():
    da = data_array(ndim=1)
    line = Line(canvas=Canvas(), data=da)
    assert np.allclose(line._line.get_xdata(), da.coords['xx'].values)
    assert np.allclose(line._line.get_ydata(), da.values)
    line.update(da * 2.5)
    assert np.allclose(line._line.get_xdata(), da.coords['xx'].values)
    assert np.allclose(line._line.get_ydata(), da.values * 2.5)


def test_line_update_with_errorbars():
    da = data_array(ndim=1, variances=True)
    line = Line(canvas=Canvas(), data=da)
    coll = line._error.get_children()[0]
    x = np.array(coll.get_segments())[:, 0, 0]
    y1 = np.array(coll.get_segments())[:, 0, 1]
    y2 = np.array(coll.get_segments())[:, 1, 1]
    assert np.allclose(x, da.coords['xx'].values)
    assert np.allclose(y1, (da.data - sc.stddevs(da.data)).values)
    assert np.allclose(y2, (da.data + sc.stddevs(da.data)).values)
    new_values = da * 2.5
    new_values.variances = da.variances
    line.update(new_values)
    coll = line._error.get_children()[0]
    y1 = np.array(coll.get_segments())[:, 0, 1]
    y2 = np.array(coll.get_segments())[:, 1, 1]
    assert np.allclose(y1, (da.data * 2.5 - sc.stddevs(da.data)).values)
    assert np.allclose(y2, (da.data * 2.5 + sc.stddevs(da.data)).values)


def test_line_datetime_binedges_with_errorbars():
    t = np.arange(
        np.datetime64('2017-03-16T20:58:17'), np.datetime64('2017-03-16T21:15:17'), 20
    )
    time = sc.array(dims=['time'], values=t)
    v = np.random.rand(time.sizes['time'] - 1)
    da = sc.DataArray(
        data=sc.array(dims=['time'], values=10 * v, variances=v), coords={'time': time}
    )
    # Checking for the locations of the errorbars is slightly strange, as the
    # get_segments() method from the LineCollection gives x values as floats, not
    # datetime, and comparing with the original data is thus difficult as the floats
    # do not seem immediately convertible to datetimes.
    # Hence, we simply check if the Line can be created without raising an exception.
    Line(canvas=Canvas(), data=da)


def test_line_color():
    da = data_array(ndim=1)
    line = Line(canvas=Canvas(), data=da)
    assert line.color == 'C0'
    assert line._line.get_color() == 'C0'
    line.color = 'C1'
    assert line.color == 'C1'
    assert line._line.get_color() == 'C1'


def test_line_color_with_errorbars():
    from matplotlib.colors import to_hex

    da = data_array(ndim=1, variances=True)
    line = Line(canvas=Canvas(), data=da)
    assert line.color == 'C0'
    assert line._line.get_color() == 'C0'
    assert to_hex(line._error.get_children()[0].get_color()) == to_hex('C0')
    line.color = 'C1'
    assert line.color == 'C1'
    assert line._line.get_color() == 'C1'
    assert to_hex(line._error.get_children()[0].get_color()) == to_hex('C1')



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import uuid
from functools import reduce

import scipp as sc


def coord_as_bin_edges(
    da: sc.DataArray, key: str, dim: str | None = None
) -> sc.Variable:
    """
    If coordinate ``key`` in DataArray ``da`` is already bin edges, return it unchanged.
    If it is midpoints, return as bin edges.

    Parameters
    ----------
    da:
        The input data array.
    key:
        The name of coordinate to potentially convert to bin edges.
    dim:
        The dimension to use for computing the bin edges (this can be different from
        the coordinate name, in the case of non-dimension coordinates).
    """
    if dim is None:
        dim = key
    x = da.coords[key]
    if x.dtype == str:
        x = sc.arange(dim, float(x.shape[0]), unit=x.unit)
    if da.coords.is_edges(key, dim=dim):
        return x
    if x.dtype in ('int32', 'int64'):
        x = x.to(dtype='float64')
    if x.sizes[dim] < 2:
        half = sc.scalar(0.5, unit=x.unit)
        return sc.concat([x[dim, 0:1] - half, x[dim, 0:1] + half], dim)
    else:
        center = sc.midpoints(x, dim=dim)
        # Note: use range of 0:1 to keep dimension dim in the slice to avoid
        # switching round dimension order in concatenate step.
        left = center[dim, 0:1] - (x[dim, 1] - x[dim, 0])
        right = center[dim, -1] + (x[dim, -1] - x[dim, -2])
        return sc.concat([left, center, right], dim)


def repeat(x: sc.Variable, dim: str, n: int) -> sc.Variable:
    """
    Equivalent of Numpy's `repeat` function: repeat the values of a variable `n` times
    along dimension `dim`.

    Parameters
    ----------
    x:
        The input variable.
    dim:
        The dimension along which to repeat.
    n:
        The number of times to repeat.
    """
    index = x.dims.index(dim) + 1
    dummy_dim = uuid.uuid4().hex
    new_dims = list(x.dims)
    new_dims.insert(index, dummy_dim)
    new_shape = list(x.shape)
    new_shape.insert(index, n)
    return x.broadcast(dims=new_dims, shape=new_shape).flatten(
        dims=[dim, dummy_dim], to=dim
    )


def maybe_number_to_variable(
    x: float | sc.Variable, unit: str | None = None
) -> sc.Variable:
    """
    If the input is a raw number, convert to a variable.
    If a unit is requested, perform the conversion to that unit before returning.
    If the input is a variable, return it (potentially converted to the requested unit).

    Parameters
    ----------
    x:
        The input int or float.
    unit:
        The unit to use for the output variable.
    """
    return (
        sc.scalar(x, unit=unit)
        if isinstance(x, int | float)
        else (x.to(unit=unit) if unit is not None else x)
    )


def maybe_variable_to_number(x: float | sc.Variable, unit=None) -> float:
    """
    If the input is a variable, return its value.
    If a unit is requested, perform the conversion to that unit first.
    If the input is a number, return it unchanged.

    Parameters
    ----------
    x:
        The input number or variable.
    unit:
        Convert the input to that unit if not ``None``.
    """
    if hasattr(x, 'unit'):
        if unit is not None:
            x = x.to(unit=unit)
        x = x.values if x.dims else x.value
    return x


def name_with_unit(var: sc.Variable, name: str | None = None) -> str:
    """
    Make a string from a variable dimension and its unit.
    The variable dimension can be overridden by specifying the ``name`` directly.

    Parameters
    ----------
    var:
        The input variable.
    name:
        The name to use to override the variable's dimension name.
    """
    text = ""
    if name is not None:
        text = name
    else:
        text = str(var.dims[-1])
    if var.unit is not None:
        unit_str = 'datetime' if var.dtype == sc.DType.datetime64 else var.unit
        text += (" " if text else "") + f"[{unit_str}]"
    return text


def value_to_string(val: float, precision: int = 3) -> str:
    """
    Convert a number to a human readable string.

    Parameters
    ----------
    val:
        The input number.
    precision:
        The number of decimal places to use for the string output.
    """
    if (not isinstance(val, float)) or (val == 0):
        text = str(val)
    elif (abs(val) >= 10.0 ** (precision + 1)) or (
        abs(val) <= 10.0 ** (-precision - 1)
    ):
        text = "{val:.{prec}e}".format(val=val, prec=precision)
    else:
        text = str(val)
        if len(text) > precision + 2 + (text[0] == '-'):
            text = "{val:.{prec}f}".format(val=val, prec=precision)
    return text


def scalar_to_string(var: sc.Variable, precision: int = 3) -> str:
    """
    Convert a scalar to a human readable string.

    Parameters
    ----------
    var:
        The input scalar.
    precision:
        The number of decimal places to use for the string output.
    """
    out = value_to_string(var.value, precision=precision)
    if (var.unit not in (None, "")) and var.dtype != sc.DType.datetime64:
        out += f" {var.unit}"
    return out


def merge_masks(masks: dict[str, sc.Variable]) -> sc.Variable:
    """
    Combine all masks into a single one using the OR operation.

    Parameters
    ----------
    masks:
        The dict holding the masks to be combined.
    """
    return reduce(lambda a, b: a | b, masks.values())


def coord_element_to_string(x: sc.Variable) -> str:
    """
    Convert a slice of a coordinate containing a single value (or two values in the
    case of a bin-edge coordinate) to a string.

    Parameters
    ----------
    x:
        The input variable (of length 1 or 2).
    """
    if not x.shape:
        return value_to_string(x.value) + (f" [{x.unit}]" if x.unit is not None else "")
    out = ':'.join([value_to_string(v) for v in x.values])
    if x.unit is not None:
        out += f" [{x.unit}]"
    return out


def make_compatible(x: sc.Variable, *, unit: str | None, dim: str | None = None):
    """
    Raise exception if the dimensions of the supplied variable do not contain the
    requested dimension.
    Then, if the variable unit differs from the requested unit, attempt a conversion.

    Parameters
    ----------
    x:
        The input variable.
    unit:
        Attempt unit conversion to this unit if the input variable has a different unit.
    dim:
        If supplied, check that the dim is found in the input variable's dimensions.
    """
    if (dim is not None) and (dim not in x.dims):
        raise sc.DimensionError(
            f'The supplied variable has dimension {x.dims} which is '
            f'incompatible with the requested dimension {dim}.'
        )
    if x.unit != unit:
        if unit is not None:
            return x.to(unit=unit, dtype=float)
        else:
            raise sc.UnitError(
                f'The supplied variable has unit {x.unit} which is '
                'incompatible with the requested unit None/<no unit>.'
            )
    return x



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import uuid

import numpy as np
import scipp as sc


def _check_ndim(data):
    if data.ndim != 1:
        raise ValueError(
            'PointCloud only accepts one dimensional data, '
            f'found {data.ndim} dimensions. You should flatten your data '
            '(using scipp.flatten) before sending it to the point cloud.'
        )


class PointCloud:
    """
    Artist to represent a three-dimensional point cloud/scatter plot.

    Parameters
    ----------
    x:
        The name of the coordinate that is to be used for the X positions.
    y:
        The name of the coordinate that is to be used for the Y positions.
    z:
        The name of the coordinate that is to be used for the Z positions.
    data:
        The initial data to create the line from.
    pixel_size:
        The size of the markers.
    opacity:
        The opacity of the points.
    """

    def __init__(
        self,
        *,
        x: str,
        y: str,
        z: str,
        data: sc.DataArray,
        pixel_size: sc.Variable | float = 1,
        opacity: float = 1,
    ):
        """
        Make a point cloud using pythreejs
        """
        import pythreejs as p3

        _check_ndim(data)
        self._data = data
        self._x = x
        self._y = y
        self._z = z
        self._id = uuid.uuid4().hex

        self._pixel_size = pixel_size
        if hasattr(self._pixel_size, 'unit'):
            if len({self._data.coords[dim].unit for dim in [x, y, z]}) > 1:
                raise ValueError(
                    f'The supplied pixel_size has unit {self._pixel_size.unit}, but '
                    'the spatial coordinates do not all have the same units. In this '
                    'case the pixel_size should just be a float with no unit.'
                )
            else:
                self._pixel_size = self._pixel_size.to(
                    dtype=float, unit=self._data.coords[x].unit
                ).value

        self.geometry = p3.BufferGeometry(
            attributes={
                'position': p3.BufferAttribute(
                    array=np.array(
                        [
                            self._data.coords[self._x].values.astype('float32'),
                            self._data.coords[self._y].values.astype('float32'),
                            self._data.coords[self._z].values.astype('float32'),
                        ]
                    ).T
                ),
                'color': p3.BufferAttribute(
                    array=np.zeros(
                        [self._data.coords[self._x].shape[0], 3], dtype='float32'
                    )
                ),
            }
        )

        # TODO: a device pixel_ratio should probably be read from a config file
        pixel_ratio = 1.0
        # Note that an additional factor of 2.5 (obtained from trial and error) seems to
        # be required to get the sizes right in the scene.
        self.material = p3.PointsMaterial(
            vertexColors='VertexColors',
            size=2.5 * self._pixel_size * pixel_ratio,
            transparent=True,
            opacity=opacity,
        )
        self.points = p3.Points(geometry=self.geometry, material=self.material)

    def set_colors(self, rgba):
        """
        Set the point cloud's rgba colors:

        Parameters
        ----------
        rgba:
            The array of rgba colors.
        """
        self.geometry.attributes["color"].array = rgba[..., :3].astype('float32')

    def update(self, new_values):
        """
        Update point cloud array with new values.

        Parameters
        ----------
        new_values:
            New data to update the point cloud values from.
        """
        _check_ndim(new_values)
        self._data = new_values

    def get_limits(self) -> tuple[sc.Variable, sc.Variable, sc.Variable]:
        """
        Get the spatial extent of all the points in the cloud.
        """
        xcoord = self._data.coords[self._x]
        ycoord = self._data.coords[self._y]
        zcoord = self._data.coords[self._z]
        half_pixel = 0.5 * self._pixel_size
        dx = sc.scalar(half_pixel, unit=xcoord.unit)
        dy = sc.scalar(half_pixel, unit=ycoord.unit)
        dz = sc.scalar(half_pixel, unit=zcoord.unit)
        return (
            sc.concat([xcoord.min() - dx, xcoord.max() + dx], dim=self._x),
            sc.concat([ycoord.min() - dy, ycoord.max() + dy], dim=self._y),
            sc.concat([zcoord.min() - dz, zcoord.max() + dz], dim=self._z),
        )

    @property
    def opacity(self):
        """
        Get the material opacity.
        """
        return self.material.opacity

    @opacity.setter
    def opacity(self, val):
        """
        Set the material opacity.
        """
        self.material.opacity = val
        self.material.depthTest = val > 0.5

    @property
    def data(self):
        """
        Get the point cloud data.
        """
        return self._data



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

import asyncio
from collections.abc import Callable


class Timer:
    """
    From:
    https://ipywidgets.readthedocs.io/en/8.0.2/examples/Widget%20Events.html#Debouncing
    """

    def __init__(self, timeout: float, callback: Callable):
        self._timeout = timeout
        self._callback = callback

    async def _job(self):
        await asyncio.sleep(self._timeout)
        self._callback()

    def start(self):
        self._task = asyncio.ensure_future(self._job())

    def cancel(self):
        self._task.cancel()


def debounce(wait: float):
    """
    Decorator that will postpone a function's
    execution until after `wait` seconds
    have elapsed since the last time it was invoked.

    From:
    https://ipywidgets.readthedocs.io/en/8.0.2/examples/Widget%20Events.html#Debouncing
    """

    def decorator(fn: Callable):
        timer = None

        def debounced(*args, **kwargs):
            nonlocal timer

            def call_it():
                fn(*args, **kwargs)

            if timer is not None:
                timer.cancel()
            timer = Timer(wait, call_it)
            timer.start()

        return debounced

    return decorator



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

from ..core import Node, View
from ..core.utils import make_compatible, name_with_unit


class GraphicalView(View):
    """
    Base class for graphical 1d and 2d views.
    It is used to represent line plots, scatter plots, and image plots (heatmaps).
    In addition to ``View``, it updates the canvas axes and labels when new data is
    supplied.
    It also verifies that the new data supplied is compatible with the existing axes
    dimensions and units.
    """

    def __init__(self, *nodes: Node):
        super().__init__(*nodes)
        self.colormapper = None

    def update(self, *args, **kwargs):
        """
        Update the view with new data by either supplying a dictionary of
        new data or by keyword arguments.
        """

        new = dict(*args, **kwargs)
        for key, new_values in new.items():
            if new_values.ndim != self._ndim:
                raise ValueError(
                    f"Expected {self._ndim} dimension(s), but got {new_values.ndim}."
                )

            coords = {}
            for i, direction in enumerate(self._dims):
                if self._dims[direction] is None:
                    self._dims[direction] = new_values.dims[i]
                coords[direction] = new_values.coords[self._dims[direction]]

            if self.canvas.empty:
                axes_units = {k: coord.unit for k, coord in coords.items()}
                axes_dtypes = {k: coord.dtype for k, coord in coords.items()}
                if 'y' in self._dims:
                    self.canvas.ylabel = name_with_unit(
                        var=coords['y'], name=self._dims['y']
                    )
                    if self._dims['y'] in self._scale:
                        self.canvas.yscale = self._scale[self._dims['y']]
                else:
                    self.canvas.ylabel = name_with_unit(var=new_values.data, name="")
                    axes_units['y'] = new_values.unit
                    axes_dtypes['y'] = new_values.dtype

                self.canvas.set_axes(
                    dims=self._dims, units=axes_units, dtypes=axes_dtypes
                )
                self.canvas.xlabel = name_with_unit(
                    var=coords['x'], name=self._dims['x']
                )
                if self.colormapper is not None:
                    self.colormapper.unit = new_values.unit
                if self._dims['x'] in self._scale:
                    self.canvas.xscale = self._scale[self._dims['x']]
            else:
                if self.colormapper is not None:
                    new_values.data = make_compatible(
                        new_values.data, unit=self.colormapper.unit
                    )
                for xy, dim in self._dims.items():
                    new_values.coords[dim] = make_compatible(
                        coords[xy], unit=self.canvas.units[xy]
                    )
                if 'y' not in self._dims:
                    new_values.data = make_compatible(
                        new_values.data, unit=self.canvas.units['y']
                    )

            if key not in self.artists:
                self.artists[key] = self.make_artist(new_values)
                if self.colormapper is not None:
                    self.colormapper[key] = self.artists[key]

            self.artists[key].update(new_values=new_values)

        if self.colormapper is not None:
            self.colormapper.update(**new)
        self.canvas.autoscale()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import numpy as np
import scipp as sc

from plopp.core.utils import coord_as_bin_edges, coord_element_to_string


def test_coord_as_bin_edges_midpoints_input():
    da = sc.DataArray(
        data=sc.arange('x', 5.0, unit='K'), coords={'x': sc.arange('x', 5.0, unit='m')}
    )
    result = coord_as_bin_edges(da, key='x')
    assert sc.identical(result, sc.linspace('x', -0.5, 4.5, num=6, unit='m'))


def test_coord_as_bin_edges_edges_input():
    x = sc.arange('x', 6.0, unit='m')
    da = sc.DataArray(data=sc.arange('x', 5.0, unit='K'), coords={'x': x})
    result = coord_as_bin_edges(da, key='x')
    assert sc.identical(result, x)


def test_coord_as_bin_edges_string_midpoints_input():
    strings = ['a', 'b', 'c', 'd', 'e']
    da = sc.DataArray(
        data=sc.arange('x', 5.0, unit='K'),
        coords={'x': sc.array(dims=['x'], values=strings, unit='s')},
    )
    result = coord_as_bin_edges(da, key='x')
    assert sc.identical(result, sc.linspace('x', -0.5, 4.5, num=6, unit='s'))


def test_coord_as_bin_edges_string_edges_input():
    strings = ['a', 'b', 'c', 'd', 'e', 'f']
    da = sc.DataArray(
        data=sc.arange('x', 5.0, unit='K'),
        coords={'x': sc.array(dims=['x'], values=strings, unit='s')},
    )
    result = coord_as_bin_edges(da, key='x')
    assert sc.identical(result, sc.arange('x', 6.0, unit='s'))


def test_coord_as_bin_edges_int_midpoints_input():
    da = sc.DataArray(
        data=sc.arange('x', 5.0, unit='K'), coords={'x': sc.arange('x', 5, unit='m')}
    )
    result = coord_as_bin_edges(da, key='x')
    assert sc.identical(result, sc.linspace('x', -0.5, 4.5, num=6, unit='m'))


def test_coord_as_bin_edges_int_edges_input():
    x = sc.arange('x', 6, unit='m')
    da = sc.DataArray(data=sc.arange('x', 5.0, unit='K'), coords={'x': x})
    result = coord_as_bin_edges(da, key='x')
    assert sc.identical(result, x)


def test_coord_as_bin_edges_non_dimension_coord():
    da = sc.DataArray(
        data=sc.arange('x', 5.0, unit='K'),
        coords={
            'x': sc.arange('x', 5.0, unit='m'),
            'y': sc.arange('x', 17.0, 22.0, unit='m'),
        },
    )
    result = coord_as_bin_edges(da, key='y', dim='x')
    assert sc.identical(result, sc.linspace('x', 16.5, 21.5, num=6, unit='m'))


def test_coord_element_to_string_float():
    assert coord_element_to_string(sc.scalar(1.2, unit='K')) == '1.2 [K]'
    assert coord_element_to_string(sc.arange('x', 2.0, unit='m')) == '0.0:1.0 [m]'


def test_coord_element_to_string_datetime():
    date = '2021-06-01T17:00:00'
    start = sc.scalar(np.datetime64(date))
    datetime = start + sc.array(dims=['time'], values=[0, 3600], unit='s')
    assert coord_element_to_string(start) == date + ' [s]'
    assert (
        coord_element_to_string(datetime)
        == '2021-06-01T17:00:00:2021-06-01T18:00:00 [s]'
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import numpy as np
import pytest
import scipp as sc

from plopp.data.testing import data_array
from plopp.plotting.common import preprocess


def test_preprocess_raises_ValueError_when_given_binned_data():
    da = sc.data.table_xyz(100).bin(x=10, y=20)
    with pytest.raises(ValueError, match='Cannot plot binned data'):
        preprocess(da)


def test_preprocess_use_non_dimension_coords():
    da = data_array(ndim=2)
    da.coords['xx2'] = 7.5 * da.coords['xx']
    da.coords['yy2'] = 3.3 * da.coords['yy']
    out = preprocess(da, coords=['xx2', 'yy2'])
    assert set(out.dims) == {'xx2', 'yy2'}
    assert out.coords['xx2'].max() == 7.5 * da.coords['xx'].max()
    assert out.coords['yy2'].max() == 3.3 * da.coords['yy'].max()


def test_preprocess_warns_when_coordinate_is_not_sorted():
    da = data_array(ndim=1)
    unsorted = sc.concat([da['xx', 20:], da['xx', :20]], dim='xx')
    with pytest.warns(
        RuntimeWarning, match='The input contains a coordinate with unsorted values'
    ):
        preprocess(unsorted)


def test_preprocess_no_warning_if_dtype_cannot_be_sorted():
    da = data_array(ndim=1)
    da.coords['vecs'] = sc.vectors(
        dims=['xx'], values=np.random.random((da.sizes['xx'], 3))
    )
    out = preprocess(da)  # no warning should be emitted
    assert 'vecs' in out.coords
    da.coords['strings'] = sc.array(
        dims=['xx'], values=list('ba' * (da.sizes['xx'] // 2))
    )
    out = preprocess(da)  # no warning should be emitted
    assert 'strings' in out.coords


@pytest.mark.parametrize(
    'dtype_and_shape',
    [
        (sc.DType.affine_transform3, (4, 4)),
        (sc.DType.linear_transform3, (3, 3)),
        (sc.DType.rotation3, (4,)),
        (sc.DType.translation3, (3,)),
        (sc.DType.vector3, (3,)),
    ],
)
def test_preprocess_raises_for_unsupported_dtype(dtype_and_shape):
    dtype, shape = dtype_and_shape
    x = np.random.random((*shape, 5))
    values = x / np.broadcast_to(
        np.linalg.norm(x, axis=tuple(range(len(shape)))), x.shape
    )
    v = sc.Variable(dims=['x'], values=values.T, dtype=dtype)
    with pytest.raises(
        TypeError,
        match=f'The input has dtype {dtype} which is not supported by Plopp',
    ):
        preprocess(v)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

import uuid
from collections.abc import Callable
from functools import partial, reduce
from typing import Any, Literal

import ipywidgets as ipw
import numpy as np
import pythreejs as p3
import scipp as sc

from ..core import Node
from ..graphics import BaseFig
from .debounce import debounce
from .style import BUTTON_LAYOUT


def _xor(x: list[sc.Variable]) -> sc.Variable:
    dim = uuid.uuid4().hex
    return sc.concat(x, dim).sum(dim) == sc.scalar(1, unit=None)


OPERATIONS = {
    'or': partial(reduce, lambda x, y: sc.logical_or(x, y)),
    'and': partial(reduce, lambda x, y: sc.logical_and(x, y)),
    'xor': _xor,
}


def select(da: sc.DataArray, s: tuple[str, sc.Variable]) -> sc.DataArray:
    return da[s]


class Clip3dTool(ipw.HBox):
    """
    A tool that provides a slider to extract a slab of points in a three-dimensional
    scatter plot, and add it to the scene as an opaque cut. The slider controls the
    position and range of the slice. When the slider is dragged, the red outline of the
    cut is moved at the same time, while the actual point cloud gets updated less
    frequently using a debounce mechanism.

    .. versionadded:: 24.04.0

    Parameters
    ----------
    limits:
        The spatial extent of the points in the 3d figure in the XYZ directions.
    direction:
        The direction normal to the slice.
    update:
        A function to update the scene.
    color:
        Color of the cut's outline.
    linewidth:
        Width of the line delineating the outline.
    """

    def __init__(
        self,
        limits: tuple[sc.Variable, sc.Variable, sc.Variable],
        direction: Literal['x', 'y', 'z'],
        update: Callable,
        color: str = 'red',
        linewidth: float = 1.5,
        border_visible: bool = True,
    ):
        self._limits = limits
        self._direction = direction
        axis = 'xyz'.index(self._direction)
        self.dim = self._limits[axis].dim
        self._unit = self._limits[axis].unit
        self.visible = True
        self._update = update
        self._border_visible = border_visible

        w_axis = 2 if self._direction == 'x' else 0
        h_axis = 2 if self._direction == 'y' else 1
        width = (self._limits[w_axis][1] - self._limits[w_axis][0]).value
        height = (self._limits[h_axis][1] - self._limits[h_axis][0]).value

        self.outlines = [
            p3.LineSegments(
                geometry=p3.EdgesGeometry(
                    p3.PlaneBufferGeometry(width=width, height=height)
                ),
                material=p3.LineBasicMaterial(color=color, linewidth=linewidth),
            ),
            p3.LineSegments(
                geometry=p3.EdgesGeometry(
                    p3.PlaneBufferGeometry(width=width, height=height)
                ),
                material=p3.LineBasicMaterial(color=color, linewidth=linewidth),
            ),
        ]
        if self._direction == 'x':
            for outline in self.outlines:
                outline.rotateY(0.5 * np.pi)
        if self._direction == 'y':
            for outline in self.outlines:
                outline.rotateX(0.5 * np.pi)

        center = [var.mean().value for var in self._limits]
        vmin = self._limits[axis][0].value
        vmax = self._limits[axis][1].value
        dx = vmax - vmin
        delta = 0.05 * dx
        self.slider = ipw.FloatRangeSlider(
            min=vmin,
            max=vmax,
            value=[center[axis] - delta, center[axis] + delta],
            step=dx * 0.01,
            description=direction.upper(),
            style={'description_width': 'initial'},
            layout={'width': '470px', 'padding': '0px'},
        )

        self.cut_visible = ipw.Button(
            icon='eye-slash',
            tooltip='Hide cut',
            layout={'width': '16px', 'padding': '0px'},
        )

        for outline, val in zip(self.outlines, self.slider.value, strict=True):
            pos = list(center)
            pos[axis] = val
            outline.position = pos
            outline.visible = self._border_visible

        self.unit_label = ipw.Label(f'[{self._unit}]')
        self.cut_visible.on_click(self.toggle)
        self.slider.observe(self.move, names='value')

        super().__init__([self.slider, ipw.Label(f'[{self._unit}]'), self.cut_visible])

    def toggle(self, owner: ipw.Button):
        """
        Toggle the visibility of the cut on and off.
        """
        self.visible = not self.visible
        for outline in self.outlines:
            outline.visible = self.visible and self._border_visible
        self.slider.disabled = not self.visible
        owner.icon = 'eye-slash' if self.visible else 'eye'
        owner.tooltip = 'Hide cut' if self.visible else 'Show cut'
        self._update()

    def toggle_border(self, value: bool):
        """
        Toggle the border visibility.
        """
        for outline in self.outlines:
            outline.visible = value
        # The call to this function comes from the parent widget, so we need to
        # remember the state of the button, so that when we toggle the visbiility of
        # the cut, the border visibility is in sync with the parent button.
        self._border_visible = value

    def move(self, value: dict[str, Any]):
        """
        Move the outline of the cut according to new position given by the slider.
        """
        # Early return if relative difference between new and old value is small.
        # This also prevents flickering of an existing cut when a new cut is added.
        if (
            np.abs(np.array(value['new']) - np.array(value['old'])).max()
            < 0.01 * self.slider.step
        ):
            return
        for outline, val in zip(self.outlines, value['new'], strict=True):
            pos = list(outline.position)
            axis = 'xyz'.index(self._direction)
            pos[axis] = val
            outline.position = pos
        self._throttled_update()

    @property
    def range(self):
        return sc.scalar(self.slider.value[0], unit=self._unit), sc.scalar(
            self.slider.value[1], unit=self._unit
        )

    @debounce(0.3)
    def _throttled_update(self):
        self._update()


class ClippingPlanes(ipw.HBox):
    """
    A widget to make clipping planes for spatial cutting (see :class:`Clip3dTool`) to
    make spatial cuts in the X, Y, and Z directions on a three-dimensional scatter plot.

    .. versionadded:: 24.04.0

    Parameters
    ----------
    fig:
        The 3d figure that contains the point clouds to be cut.
    """

    def __init__(self, fig: BaseFig):
        self._view = fig._view
        self._limits = self._view.get_limits()
        self.cuts = []
        self._operation = 'or'

        self.tabs = ipw.Tab(layout={'width': '550px'})
        self._original_nodes = list(self._view.graph_nodes.values())
        self._nodes = {}

        self.add_cut_label = ipw.Label('Add cut:')
        layout = {'width': '45px', 'padding': '0px 0px 0px 0px'}
        self.add_x_cut = ipw.Button(
            description='X',
            icon='plus',
            tooltip='Add X cut',
            layout=layout,
        )
        self.add_y_cut = ipw.Button(
            description='Y',
            icon='plus',
            tooltip='Add Y cut',
            layout=layout,
        )
        self.add_z_cut = ipw.Button(
            description='Z',
            icon='plus',
            tooltip='Add Z cut',
            layout=layout,
        )
        self.add_x_cut.on_click(lambda _: self._add_cut('x'))
        self.add_y_cut.on_click(lambda _: self._add_cut('y'))
        self.add_z_cut.on_click(lambda _: self._add_cut('z'))

        self.opacity = ipw.BoundedFloatText(
            min=0,
            max=0.5,
            step=0.01,
            disabled=True,
            value=0.03,
            description='Opacity:',
            tooltip='Set the opacity of the background',
            style={'description_width': 'initial'},
            layout={'width': '142px', 'padding': '0px 0px 0px 0px'},
        )
        self.opacity.observe(self._set_opacity, names='value')

        self.cut_borders_visibility = ipw.ToggleButton(
            value=True,
            disabled=True,
            icon='border-style',
            tooltip='Toggle visibility of the borders of the cuts',
            **BUTTON_LAYOUT,
        )
        self.cut_borders_visibility.observe(
            self.toggle_border_visibility, names='value'
        )

        self.cut_operation = ipw.Dropdown(
            options=['OR', 'AND', 'XOR'],
            value='OR',
            disabled=True,
            tooltip='Operation to combine multiple cuts',
            layout={'width': '60px', 'padding': '0px 0px 0px 0px'},
        )
        self.cut_operation.observe(self.change_operation, names='value')

        self.delete_cut = ipw.Button(
            tooltip='Delete cut',
            icon='trash',
            disabled=True,
            **BUTTON_LAYOUT,
        )
        self.delete_cut.on_click(self._remove_cut)

        super().__init__(
            [
                self.tabs,
                ipw.VBox(
                    [
                        ipw.HBox([self.add_x_cut, self.add_y_cut, self.add_z_cut]),
                        self.opacity,
                        ipw.HBox(
                            [
                                self.cut_borders_visibility,
                                self.cut_operation,
                                self.delete_cut,
                            ]
                        ),
                    ]
                ),
            ]
        )

        self.layout.display = 'none'

    def _add_cut(self, direction: Literal['x', 'y', 'z']):
        """
        Add a cut in the specified direction.
        """
        cut = Clip3dTool(
            direction=direction,
            limits=self._limits,
            update=self.update_state,
            border_visible=self.cut_borders_visibility.value,
        )
        self._view.canvas.add(cut.outlines)
        self.cuts.append(cut)
        self.tabs.children = [*self.tabs.children, cut]
        self.tabs.selected_index = len(self.cuts) - 1
        self.update_controls()
        self.update_state()

    def _remove_cut(self, _):
        cut = self.cuts.pop(self.tabs.selected_index)
        self._view.canvas.remove(cut.outlines)
        self.tabs.children = self.cuts
        self.update_state()
        self.update_controls()

    def update_controls(self):
        """
        If there are no active cuts, disable the controls.
        If there is at least one cut, set the opacity of the original children
        (not the cuts) to a low value, otherwise set it back to 1.
        """
        at_least_one_cut = any(1 for cut in self.cuts if cut.visible)
        self.delete_cut.disabled = not at_least_one_cut
        self.cut_borders_visibility.disabled = not at_least_one_cut
        self.cut_operation.disabled = not at_least_one_cut
        self.tabs.titles = [cut._direction.upper() for cut in self.cuts]
        self.opacity.disabled = not at_least_one_cut
        opacity = self.opacity.value if at_least_one_cut else 1.0
        self._set_opacity({'new': opacity})

    def _set_opacity(self, change: dict[str, Any]):
        """
        Set the opacity of the original point clouds in the figure, not the cuts.
        """
        self._view.set_opacity(change['new'])

    def toggle_visibility(self):
        """
        Toggle the visibility of the control buttons for making the cuts.
        """
        self.layout.display = None if self.layout.display == 'none' else 'none'

    def toggle_border_visibility(self, change: dict[str, Any]):
        """
        Toggle the visibility of the borders of the cuts.
        """
        for cut in self.cuts:
            cut.toggle_border(change['new'])

    def change_operation(self, change: dict[str, Any]):
        """
        Change the operation to combine multiple cuts.
        """
        self._operation = change['new'].lower()
        self.update_state()

    def update_state(self):
        """
        Update the state, combining all the active cuts, using the selected binary
        operation. The resulting selection is then used to either create or update a
        second point cloud which is included in the scene.
        The original point cloud is then set to be semi-transparent.
        When the position/range of a cut is changed, this function is called via a
        debounce mechanism to avoid updating the cloud too often. Only the outlines of
        the cuts are moved in real time, which is cheap.
        """
        for nodes in self._nodes.values():
            self._view.remove(nodes['slice'].id)
            nodes['slice'].remove()
        self._nodes.clear()

        visible_cuts = [cut for cut in self.cuts if cut.visible]
        if not visible_cuts:
            return

        for n in self._original_nodes:
            da = n.request_data()
            selections = []
            for cut in visible_cuts:
                xmin, xmax = cut.range
                selections.append(
                    (da.coords[cut.dim] >= xmin) & (da.coords[cut.dim] < xmax)
                )
            selection = OPERATIONS[self._operation](selections)
            if selection.sum().value > 0:
                if n.id not in self._nodes:
                    select_node = Node(selection)
                    self._nodes[n.id] = {
                        'select': select_node,
                        'slice': Node(lambda da, s: da[s], da=n, s=select_node),
                    }
                    self._nodes[n.id]['slice'].add_view(self._view)
                else:
                    self._nodes[n.id]['select'].func = lambda: selection  # noqa: B023
                self._nodes[n.id]['select'].notify_children("")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import warnings
from collections.abc import Callable
from typing import Any

import numpy as np
import scipp as sc

from .. import backends
from ..core import Node
from ..core.typing import Plottable, PlottableMulti


def require_interactive_backend(func: str):
    """
    Raise an error if the current backend in use is non-interactive.
    """
    if not backends.is_interactive():
        raise RuntimeError(
            f"The {func} can only be used with an interactive backend "
            "backend. Use `%matplotlib widget` at the start of your "
            "notebook."
        )


def is_pandas_series(obj: Any) -> bool:
    """
    Check if an object is a pandas series.
    """
    try:
        import pandas as pd
    except ImportError:
        return False
    return isinstance(obj, pd.Series)


def from_compatible_lib(obj: Any) -> Any:
    """
    Convert from a compatible library, if possible.
    """
    if 'pandas' in str(type(obj)):
        return sc.compat.from_pandas(obj)
    if 'xarray' in str(type(obj)):
        return sc.compat.from_xarray(obj)
    return obj


def _maybe_to_variable(obj: Plottable | list) -> Plottable:
    """
    Attempt to convert the input to a Variable.
    If the input is either a list or a numpy array, it will be converted.
    Otherwise, the input will be returned unchanged.
    """
    out = obj
    if isinstance(out, list):
        out = np.array(out)
    if isinstance(out, np.ndarray):
        dims = [f"axis-{i}" for i in range(len(out.shape))]
        out = sc.Variable(dims=dims, values=out)
    return out


def to_variable(obj) -> sc.Variable:
    """
    Convert an input to a Variable. If the object returned by the conversion is not a
    Variable, raise an error.

    Parameters
    ----------
    obj:
        The input object to be converted.
    """
    out = _maybe_to_variable(obj)
    if not isinstance(out, sc.Variable):
        raise TypeError(f"Cannot convert input of type {type(obj)} to a Variable.")
    return out


def to_data_array(
    obj: Plottable | list,
) -> sc.DataArray:
    """
    Convert an input to a DataArray, potentially adding fake coordinates if they are
    missing.

    Parameters
    ----------
    obj:
        The input object to be converted.
    """
    out = _maybe_to_variable(obj)
    if isinstance(out, sc.Variable):
        out = sc.DataArray(data=out)
    out = from_compatible_lib(out)
    if not isinstance(out, sc.DataArray):
        raise TypeError(f"Cannot convert input of type {type(obj)} to a DataArray.")
    out = out.copy(deep=False)
    for dim, size in out.sizes.items():
        if dim not in out.coords:
            out.coords[dim] = sc.arange(dim, size, unit=None)
    return out


def _check_size(da: sc.DataArray):
    """
    Prevent slow figure rendering by raising an error if the data array exceeds a
    default size.
    """
    limits = {1: 1_000_000, 2: 2500 * 2500}
    if da.ndim not in limits:
        raise ValueError("plot can only handle 1d and 2d data.")
    if np.prod(da.shape) > limits[da.ndim]:
        raise ValueError(
            f"Plotting data of size {da.shape} may take very long or use "
            "an excessive amount of memory. This is therefore disabled by "
            "default. To bypass this check, use `ignore_size=True`."
        )


def check_not_binned(da: sc.DataArray):
    """
    Plopp cannot plot binned data.
    This function will raise an error if the input data is binned.

    Parameters
    ----------
    da:
        The data array to be plotted.
    """
    if da.bins is not None:
        params = ', '.join([f'{dim}=100' for dim in da.dims])
        raise ValueError(
            "Cannot plot binned data, it must be histogrammed first, "
            f"e.g., using ``da.hist()`` or da.hist({params})``."
            "See https://scipp.github.io/generated/functions/scipp.hist.html for "
            "more details."
        )


def check_allowed_dtypes(da: sc.DataArray):
    """
    Currently, Plopp cannot plot data that contains vector and matrix dtypes.
    This function will raise an error if the input data type is not supported.

    Parameters
    ----------
    da:
        The data array to be plotted.
    """
    if da.shape != da.values.shape:
        raise TypeError(
            f'The input has dtype {da.dtype} which is not supported by Plopp.'
        )


def _all_dims_sorted(var, order='ascending'):
    return all(sc.allsorted(var, dim, order=order) for dim in var.dims)


def preprocess(
    obj: Plottable | list,
    name: str | None = None,
    ignore_size: bool = False,
    coords: list[str] | None = None,
) -> sc.DataArray:
    """
    Pre-process input data for plotting.
    This involves:

      - converting the input to a data array
      - filling in missing dimension coords if needed
      - renaming dimensions if non-dimension coordinates are to be used

    Parameters
    ----------
    obj:
        The input object that will be plotted.
    name:
        Override the input's name (when a dict-like structure is to be plotted).
    ignore_size:
        Do not perform a size check on the object before plotting it.
    coords:
        If supplied, use these coords instead of the input's dimension coordinates.
    """
    out = to_data_array(obj)
    check_not_binned(out)
    check_allowed_dtypes(out)
    if name is not None:
        out.name = name
    if not ignore_size:
        _check_size(out)
    if coords is not None:
        renamed_dims = {}
        if isinstance(coords, str):
            coords = [coords]
        for dim in coords:
            underlying = out.coords[dim].dims[-1]
            renamed_dims[underlying] = dim
        out = out.rename_dims(**renamed_dims)
    for n, coord in out.coords.items():
        if (coord.ndim == 0) or (n not in out.dims):
            continue
        try:
            if not (
                _all_dims_sorted(coord, order='ascending')
                or _all_dims_sorted(coord, order='descending')
            ):
                warnings.warn(
                    'The input contains a coordinate with unsorted values '
                    f'({n}). The results may be unpredictable. '
                    'Coordinates can be sorted using '
                    '`scipp.sort(data, dim="to_be_sorted", order="ascending")`.',
                    RuntimeWarning,
                    stacklevel=2,
                )
        except sc.DTypeError:
            pass
    return out


def input_to_nodes(obj: PlottableMulti, processor: Callable) -> list[Node]:
    """
    Convert an input or dict of inputs to a list of nodes that provide pre-processed
    data.

    Parameters
    ----------
    obj:
        The input(s) to be converted to nodes.
    processor:
        The function that will be applied to each input to convert it to a node.
    """
    if hasattr(obj, 'items') and not is_pandas_series(obj):
        to_nodes = obj.items()
    else:
        to_nodes = [(None, obj)]
    nodes = [Node(processor, inp, name=name) for name, inp in to_nodes]
    for node in nodes:
        if hasattr(processor, 'func'):
            node.name = processor.func.__name__
        else:
            node.name = 'Preprocess data'
    return nodes


def raise_multiple_inputs_for_2d_plot_error(origin):
    """
    Raise an error if the user tries to plot multiple 2d data entries.

    Parameters
    ----------
    origin:
        The name of the function that called this function.
    """
    raise ValueError(
        f'The {origin} function can only plot a single 2d data entry. If you want '
        'to create multiple figures, see the documentation on subplots at '
        'https://scipp.github.io/plopp/customization/subplots.html. If you '
        'want to plot two images onto the same axes, use the lower-level '
        'plopp.imagefigure function.'
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)


from ..core.typing import FigureLike, Plottable
from .common import require_interactive_backend
from .slicer import Slicer


def superplot(
    obj: Plottable,
    keep: str | None = None,
    **kwargs,
) -> FigureLike:
    """
    Plot a multi-dimensional object as a one-dimensional line, slicing all but one
    dimension. This will produce one slider per sliced dimension, below the figure.
    In addition, a tool for saving the currently displayed line is added on the right
    hand side of the figure.

    Parameters
    ----------
    obj:
        The object to be plotted.
    keep:
        The single dimension to be kept, all remaining dimensions will be sliced.
        This should be a single string. If no dim is provided, the last/inner dim will
        be kept.
    **kwargs:
        See :py:func:`plopp.plot` for the full list of line customization arguments.

    Returns
    -------
    :
        A :class:`widgets.Box` which will contain a :class:`graphics.FigLine`, slider
        widgets and a tool to save/delete lines.
    """
    from ..widgets import LineSaveTool

    require_interactive_backend('superplot')
    slicer = Slicer(obj, keep=keep, **kwargs)
    slicer.figure.right_bar.add(
        LineSaveTool(
            data_node=slicer.slice_nodes[0],
            slider_node=slicer.slider_node,
            fig=slicer.figure,
        )
    )
    return slicer.figure



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from functools import partial

from . import factory as fa

testing_dim_list = ['xx', 'yy', 'zz', 'time', 'temperature']

variable = partial(fa.variable, dim_list=testing_dim_list)
data_array = partial(fa.data_array, dim_list=testing_dim_list)
dataset = partial(fa.dataset, dim_list=testing_dim_list)
scatter = fa.scatter



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import os
import tempfile

import numpy as np
import pytest
import scipp as sc

import plopp as pp
from plopp.data.testing import data_array, variable


def test_plot_ndarray():
    pp.plot(np.arange(50.0).reshape(5, 10))


def test_plot_ndarray_int():
    pp.plot(np.arange(50).reshape(5, 10))


def test_plot_variable():
    pp.plot(variable(ndim=2))


def test_plot_data_array():
    pp.plot(data_array(ndim=2))


def test_plot_from_node():
    da = data_array(ndim=2)
    pp.plot(pp.Node(da))


def test_plot_data_array_2d_with_one_missing_coord_and_binedges():
    da = sc.data.table_xyz(100).bin(x=10, y=12).bins.mean()
    del da.coords['x']
    pp.plot(da)


def test_plot_2d_coord():
    da = data_array(ndim=2, ragged=True)
    pp.plot(da)
    pp.plot(da.transpose())


def test_plot_2d_coord_with_mask():
    da = data_array(ndim=2, ragged=True)
    da.masks['negative'] = da.data < sc.scalar(0, unit='m/s')
    pp.plot(da)


def test_kwarg_cmap():
    da = data_array(ndim=2)
    p = pp.plot(da, cmap='magma')
    assert p._view.colormapper.cmap.name == 'magma'


def test_kwarg_scale_2d():
    da = data_array(ndim=2)
    p = pp.plot(da, scale={'xx': 'log', 'yy': 'log'})
    assert p.canvas.ax.get_xscale() == 'log'
    assert p.canvas.ax.get_yscale() == 'log'


def test_raises_ValueError_when_given_binned_data():
    da = sc.data.table_xyz(100).bin(x=10, y=20)
    with pytest.raises(ValueError, match='Cannot plot binned data'):
        pp.plot(da)


def test_use_non_dimension_coords():
    da = data_array(ndim=2, binedges=True)
    da.coords['xx2'] = 7.5 * da.coords['xx']
    da.coords['yy2'] = 3.3 * da.coords['yy']
    p = pp.plot(da, coords=['xx2', 'yy2'])
    assert p.canvas.dims['x'] == 'xx2'
    assert p.canvas.dims['y'] == 'yy2'
    assert p.canvas.xmax == 7.5 * da.coords['xx'].max().value
    assert p.canvas.ymax == 3.3 * da.coords['yy'].max().value


@pytest.mark.parametrize('ext', ['jpg', 'png', 'pdf', 'svg'])
def test_save_to_disk_2d(ext):
    da = data_array(ndim=2)
    fig = pp.plot(da)
    with tempfile.TemporaryDirectory() as path:
        fname = os.path.join(path, f'plopp_fig2d.{ext}')
        fig.save(filename=fname)
        assert os.path.isfile(fname)


def test_save_to_disk_with_bad_extension_raises():
    da = data_array(ndim=2)
    fig = pp.plot(da)
    with pytest.raises(ValueError, match='txt'):
        fig.save(filename='plopp_fig2d.txt')


def test_plot_raises_with_multiple_2d_inputs():
    a = data_array(ndim=2)
    b = 3.3 * a
    with pytest.raises(
        ValueError, match='The plot function can only plot a single 2d data entry'
    ):
        pp.plot({'a': a, 'b': b})


def test_plot_xarray_data_array_2d():
    import xarray as xr

    N = 50
    M = 40
    data = np.random.random([M, N])
    time = np.arange(float(N))
    space = np.arange(float(M))
    da = xr.DataArray(
        data, coords={'space': space, 'time': time}, dims=['space', 'time']
    )
    p = pp.plot(da)
    assert p.canvas.dims['x'] == 'time'
    assert p.canvas.dims['y'] == 'space'
    assert p.canvas.units['x'] == 'dimensionless'
    assert p.canvas.units['y'] == 'dimensionless'


def test_plot_2d_ignores_masked_data_for_colorbar_range():
    da = data_array(ndim=2)
    da['xx', 10]['yy', 10].values = 100
    da.masks['m'] = da.data > sc.scalar(5.0, unit='m/s')
    p = pp.plot(da)
    assert p._view.colormapper.vmax < 100


def test_plot_2d_includes_masked_data_in_horizontal_range():
    da = data_array(ndim=2)
    da.masks['left'] = da.coords['xx'] < sc.scalar(5.0, unit='m')
    da.masks['right'] = da.coords['xx'] > sc.scalar(30.0, unit='m')
    p = pp.plot(da)
    assert p.canvas.xmin < 1.0
    assert p.canvas.xmax > 40.0


def test_plot_2d_includes_masked_data_in_vertical_range():
    da = data_array(ndim=2)
    da.masks['bottom'] = da.coords['yy'] < sc.scalar(5.0, unit='m')
    da.masks['top'] = da.coords['yy'] > sc.scalar(20.0, unit='m')
    p = pp.plot(da)
    assert p.canvas.ymin < 1.0
    assert p.canvas.ymax > 30.0


def test_plot_2d_datetime_coord():
    t = np.arange(
        np.datetime64('2017-03-16T20:58:17'), np.datetime64('2017-03-16T21:15:17'), 20
    )
    time = sc.array(dims=['time'], values=t)
    z = sc.arange('z', 50.0, unit='m')
    v = np.random.random(z.shape + time.shape)
    da = sc.DataArray(
        data=sc.array(dims=['z', 'time'], values=10 * v, variances=v),
        coords={'time': time, 'z': z},
    )
    pp.plot(da)


def test_plot_2d_datetime_coord_binedges():
    t = np.arange(
        np.datetime64('2017-03-16T20:58:17'), np.datetime64('2017-03-16T21:15:17'), 20
    )
    time = sc.array(dims=['time'], values=t)
    z = sc.arange('z', 50.0, unit='m')
    v = np.random.random(z[:-1].shape + time[:-1].shape)
    da = sc.DataArray(
        data=sc.array(dims=['z', 'time'], values=10 * v, variances=v),
        coords={'time': time, 'z': z},
    )
    pp.plot(da)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)


from plopp.backends.matplotlib.tiled import Tiled
from plopp.data.testing import data_array


def test_side_by_side():
    da1 = data_array(ndim=1)
    da2 = data_array(ndim=2)
    tiled = Tiled(nrows=1, ncols=2)
    tiled[0, 0] = da1.plot()
    tiled[0, 1] = da2.plot()
    assert tiled.nrows == 1
    assert tiled.ncols == 2


def test_top_bottom():
    da1 = data_array(ndim=1)
    da2 = data_array(ndim=2)
    tiled = Tiled(nrows=2, ncols=1)
    tiled[0, 0] = da1.plot()
    tiled[1, 0] = da2.plot()
    assert tiled.nrows == 2
    assert tiled.ncols == 1


def test_grid():
    da1 = data_array(ndim=1)
    da2 = data_array(ndim=2)
    tiled = Tiled(nrows=2, ncols=2)
    tiled[0, 0] = da1.plot()
    tiled[0, 1] = da2.plot()
    tiled[1, 0] = da1.plot()
    tiled[1, 1] = da2.plot()
    assert tiled.nrows == 2
    assert tiled.ncols == 2


def test_range():
    da1 = data_array(ndim=1)
    da2 = data_array(ndim=2)
    tiled = Tiled(nrows=1, ncols=3)
    tiled[0, :2] = da1.plot()
    tiled[0, 2] = da2.plot()
    assert tiled.nrows == 1
    assert tiled.ncols == 3


def test_add():
    da1 = data_array(ndim=1)
    da2 = data_array(ndim=2)
    da3 = data_array(ndim=1, binedges=True)
    da4 = data_array(ndim=2, binedges=True)
    tiled1 = Tiled(nrows=1, ncols=2)
    tiled1[0, 0] = da1.plot()
    tiled1[0, 1] = da2.plot()
    tiled2 = Tiled(nrows=1, ncols=2)
    tiled2[0, 0] = da3.plot()
    tiled2[0, 1] = da4.plot()
    tiled = tiled1 + tiled2
    assert tiled.nrows == 1
    assert tiled.ncols == 4


def test_add_tiled_figure():
    da1 = data_array(ndim=1)
    da2 = data_array(ndim=2)
    da3 = data_array(ndim=1, binedges=True)
    tiled1 = Tiled(nrows=1, ncols=2)
    tiled1[0, 0] = da1.plot()
    tiled1[0, 1] = da2.plot()
    tiled = tiled1 + da3.plot()
    assert tiled.nrows == 1
    assert tiled.ncols == 3


def test_add_figure_tiled():
    da1 = data_array(ndim=1)
    da2 = data_array(ndim=2)
    da3 = data_array(ndim=1, binedges=True)
    tiled1 = Tiled(nrows=1, ncols=2)
    tiled1[0, 0] = da1.plot()
    tiled1[0, 1] = da2.plot()
    tiled = da3.plot() + tiled1
    assert tiled.nrows == 1
    assert tiled.ncols == 3


def test_divide():
    da1 = data_array(ndim=1)
    da2 = data_array(ndim=2)
    da3 = data_array(ndim=1, binedges=True)
    da4 = data_array(ndim=2, binedges=True)
    tiled1 = Tiled(nrows=1, ncols=2)
    tiled1[0, 0] = da1.plot()
    tiled1[0, 1] = da2.plot()
    tiled2 = Tiled(nrows=1, ncols=2)
    tiled2[0, 0] = da3.plot()
    tiled2[0, 1] = da4.plot()
    tiled = tiled1 / tiled2
    assert tiled.nrows == 2
    assert tiled.ncols == 2


def test_divide_tiled_figure():
    da1 = data_array(ndim=1)
    da2 = data_array(ndim=2)
    da3 = data_array(ndim=1, binedges=True)
    da4 = data_array(ndim=2, binedges=True)
    tiled1 = Tiled(nrows=1, ncols=3)
    tiled1[0, 0] = da1.plot()
    tiled1[0, 1] = da2.plot()
    tiled1[0, 2] = da3.plot()
    tiled = tiled1 / da4.plot()
    assert tiled.nrows == 2
    assert tiled.ncols == 3


def test_divide_figure_tiled():
    da1 = data_array(ndim=1)
    da2 = data_array(ndim=2)
    da3 = data_array(ndim=1, binedges=True)
    da4 = data_array(ndim=2, binedges=True)
    tiled1 = Tiled(nrows=1, ncols=3)
    tiled1[0, 0] = da1.plot()
    tiled1[0, 1] = da2.plot()
    tiled1[0, 2] = da3.plot()
    tiled = da4.plot() / tiled1
    assert tiled.nrows == 2
    assert tiled.ncols == 3


def test_figure_add():
    da1 = data_array(ndim=1)
    da2 = data_array(ndim=2)
    tiled = da1.plot() + da2.plot()
    assert tiled.nrows == 1
    assert tiled.ncols == 2


def test_figure_divide():
    da1 = data_array(ndim=1)
    da2 = data_array(ndim=2)
    tiled = da1.plot() / da2.plot()
    assert tiled.nrows == 2
    assert tiled.ncols == 1


def test_figure_operators():
    da1 = data_array(ndim=1)
    da2 = data_array(ndim=2)
    da3 = data_array(ndim=1, binedges=True)
    da4 = data_array(ndim=2, binedges=True)
    tiled = (da1.plot() + da2.plot()) / (da3.plot() + da4.plot())
    assert tiled.nrows == 2
    assert tiled.ncols == 2


def test_tiled_keeps_figure_kwargs():
    da1 = data_array(ndim=1)
    da2 = data_array(ndim=1) * 3.3
    p1 = da1.plot(grid=True, title="My Title", vmin=-2.3, vmax=10)
    p2 = da2.plot(norm='log')
    tiled = p1 + p2
    assert tiled[0, 0].canvas.grid
    assert tiled[0, 0].canvas.title == "My Title"
    assert tiled[0, 0].canvas.ymin == -2.3
    assert tiled[0, 0].canvas.ymax == 10
    assert tiled[0, 1].canvas.yscale == 'log'


def test_tiled_keeps_figure_props():
    da1 = data_array(ndim=1)
    da2 = data_array(ndim=1) * 3.3
    p1 = da1.plot()
    p2 = da2.plot()
    p1.canvas.logy()
    p2.canvas.logx()
    tiled = p1 + p2
    assert tiled[0, 0].canvas.xscale == 'linear'
    assert tiled[0, 0].canvas.yscale == 'log'
    assert tiled[0, 1].canvas.xscale == 'log'
    assert tiled[0, 1].canvas.yscale == 'linear'



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from typing import Any

from ipywidgets import VBox

from ..graphics import ColorMapper
from . import tools


class Toolbar(VBox):
    """
    Custom toolbar to control interactive figures.

    Parameters
    ----------
    tools:
        Dictionary of tools to populate the toolbar.
    """

    def __init__(self, tools: dict[str, Any] | None = None):
        self.tools = {}
        if tools is not None:
            for key, tool in tools.items():
                if isinstance(tool.callback, dict):
                    for name, cb in tool.callback.items():
                        setattr(self, name, cb)
                else:
                    setattr(self, key, tool.callback)
                self.tools[key] = tool

        super().__init__()
        self._update_children()

    def __getitem__(self, key: str) -> Any:
        return self.tools[key]

    def __setitem__(self, key: str, tool: Any):
        self.tools[key] = tool
        self._update_children()

    def __delitem__(self, key: str):
        del self.tools[key]
        self._update_children()

    def _update_children(self):
        self.children = list(self.tools.values())


def make_toolbar_canvas2d(
    canvas: Any, colormapper: ColorMapper | None = None
) -> Toolbar:
    """
    Create a toolbar for a 2D canvas.
    If the colormapper is defined, add a button to toggle the norm of the colormapper.

    Parameters
    ----------
    canvas:
        The 2D canvas to operate on.
    colormapper:
        The colormapper which controls the colors of the artists in the canvas (for
        image plots).
    """
    tool_list = {
        'home': tools.HomeTool(canvas.autoscale),
        'panzoom': tools.PanZoomTool(canvas.panzoom),
        'logx': tools.LogxTool(canvas.logx, value=canvas.xscale == 'log'),
        'logy': tools.LogyTool(canvas.logy, value=canvas.yscale == 'log'),
    }
    if colormapper is not None:
        tool_list['lognorm'] = tools.LogNormTool(
            colormapper.toggle_norm, value=colormapper.norm == 'log'
        )
    tool_list['save'] = tools.SaveTool(canvas.download_figure)
    return Toolbar(tools=tool_list)


def make_toolbar_canvas3d(
    canvas: Any, colormapper: ColorMapper | None = None
) -> Toolbar:
    """
    Create a toolbar for a 3D canvas.
    If the colormapper is defined, add a button to toggle the norm of the colormapper.

    Parameters
    ----------
    canvas:
        The 3D canvas to operate on.
    colormapper:
        The colormapper which controls the colors of the artists in the canvas.
    """
    tool_list = {
        'home': tools.HomeTool(canvas.home),
        'camerax': tools.CameraTool(
            canvas.camera_x_normal,
            description='X',
            tooltip='Camera to X normal. ' 'Click twice to flip the view direction.',
        ),
        'cameray': tools.CameraTool(
            canvas.camera_y_normal,
            description='Y',
            tooltip='Camera to Y normal. ' 'Click twice to flip the view direction.',
        ),
        'cameraz': tools.CameraTool(
            canvas.camera_z_normal,
            description='Z',
            tooltip='Camera to Z normal. ' 'Click twice to flip the view direction.',
        ),
    }
    if colormapper is not None:
        tool_list['lognorm'] = tools.LogNormTool(
            colormapper.toggle_norm, value=colormapper.norm == 'log'
        )
    tool_list.update(
        {
            'box': tools.OutlineTool(canvas.toggle_outline),
            'axes': tools.AxesTool(canvas.toggle_axes3d),
            'size': tools.PlusMinusTool(
                plus={'callback': canvas.bigger, 'tooltip': 'Increase canvas size'},
                minus={'callback': canvas.smaller, 'tooltip': 'Decrease canvas size'},
            ),
        }
    )
    return Toolbar(tools=tool_list)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from dataclasses import dataclass

import numpy as np
import scipp as sc
from matplotlib.colors import LogNorm, Normalize

from plopp.data.testing import data_array
from plopp.graphics.colormapper import ColorMapper


class DummyChild:
    def __init__(self, data):
        self._data = data
        self._colors = None

    def set_colors(self, colors):
        self._colors = colors

    def update(self, data):
        self._data = data

    @property
    def data(self):
        return self._data


def test_creation():
    mapper = ColorMapper(
        cmap='magma',
        mask_cmap='jet',
        norm='linear',
        vmin=sc.scalar(1, unit='K'),
        vmax=sc.scalar(10, unit='K'),
    )
    assert mapper.cmap.name == 'magma'
    assert mapper.mask_cmap.name == 'jet'
    assert mapper.norm == 'linear'
    assert sc.identical(mapper.user_vmin, sc.scalar(1, unit='K'))
    assert sc.identical(mapper.user_vmax, sc.scalar(10, unit='K'))


def test_norm():
    mapper = ColorMapper(norm='log')
    assert mapper.norm == 'log'


def test_autoscale():
    da = data_array(ndim=2, unit='K')
    mapper = ColorMapper()
    artist = DummyChild(da)
    mapper['data'] = artist

    mapper.update(data=da)
    assert mapper.vmin == da.min().value
    assert mapper.vmax == da.max().value

    const = 2.3
    artist.update(da * const)
    mapper.update(data=da * const)
    assert mapper.vmin == (da.min() * const).value
    assert mapper.vmax == (da.max() * const).value


def test_auto_rescale_limits_can_shrink():
    da = data_array(ndim=2, unit='K')
    mapper = ColorMapper(autoscale='auto')
    artist = DummyChild(da)
    mapper['data'] = artist

    mapper.update(data=da)
    assert mapper.vmin == da.min().value
    assert mapper.vmax == da.max().value

    const = 0.5
    artist.update(da * const)
    mapper.update(data=da * const)
    assert mapper.vmin == da.min().value * const
    assert mapper.vmax == da.max().value * const


def test_grow_rescale_limits_do_not_shrink():
    da = data_array(ndim=2, unit='K')
    mapper = ColorMapper(autoscale='grow')
    artist = DummyChild(da)
    mapper['data'] = artist

    mapper.update(data=da)
    assert mapper.vmin == da.min().value
    assert mapper.vmax == da.max().value

    const = 0.5
    artist.update(da * const)
    mapper.update(data=da * const)
    assert mapper.vmin == da.min().value
    assert mapper.vmax == da.max().value


def test_correct_normalizer_limits():
    da = sc.DataArray(data=sc.array(dims=['y', 'x'], values=[[1, 2], [3, 4]]))
    mapper = ColorMapper()
    artist = DummyChild(da)
    mapper['data'] = artist
    mapper.update(data=da)
    assert mapper.vmin == da.min().value
    assert mapper.vmax == da.max().value
    # The normalizer initially has limits [0, 1].
    # In Matplotlib, if we set the normalizer vmin value (1) equal to the current vmax,
    # it will silently set it to something smaller, e.g. 0.9.
    # Our implementation needs to work around this.
    assert mapper.normalizer.vmin == da.min().value
    assert mapper.normalizer.vmax == da.max().value


def test_vmin_vmax():
    da = data_array(ndim=2, unit='K') * 100.0
    vmin = sc.scalar(-0.1, unit='K')
    vmax = sc.scalar(3.5, unit='K')
    mapper = ColorMapper(vmin=vmin, vmax=vmax)
    artist = DummyChild(da)
    mapper['data'] = artist
    mapper.update(data=da)
    assert mapper.user_vmin == vmin.value
    assert mapper.user_vmax == vmax.value
    assert mapper.vmin == vmin.value
    assert mapper.vmax == vmax.value


def test_vmin_vmax_no_variable():
    da = data_array(ndim=2, unit='K') * 100.0
    vmin = -0.1
    vmax = 3.5
    mapper = ColorMapper(vmin=vmin, vmax=vmax)
    artist = DummyChild(da)
    mapper['data'] = artist
    mapper.update(data=da)
    assert mapper.user_vmin == vmin
    assert mapper.user_vmax == vmax
    assert mapper.vmin == vmin
    assert mapper.vmax == vmax


def test_toggle_norm():
    mapper = ColorMapper()
    da = data_array(ndim=2, unit='K')
    mapper['child1'] = DummyChild(da)
    mapper.update(child1=da)
    assert mapper.norm == 'linear'
    assert isinstance(mapper.normalizer, Normalize)
    assert mapper.vmin == da.min().value
    assert mapper.vmax == da.max().value

    mapper.toggle_norm()
    assert mapper.norm == 'log'
    assert isinstance(mapper.normalizer, LogNorm)
    assert mapper.vmin > 0
    assert mapper.vmax == da.max().value


def test_update_changes_limits():
    da = data_array(ndim=2, unit='K')
    mapper = ColorMapper()
    artist = DummyChild(da)
    mapper['data'] = artist

    mapper.update(data=da)
    assert mapper.normalizer.vmin == da.min().value
    assert mapper.normalizer.vmax == da.max().value

    const = 2.3
    artist.update(da * const)
    mapper.update(data=da * const)
    assert mapper.normalizer.vmin == (da.min() * const).value
    assert mapper.normalizer.vmax == (da.max() * const).value


def test_rgba():
    da = data_array(ndim=2, unit='K')
    mapper = ColorMapper()
    colors = mapper.rgba(da)
    assert colors.shape == (*da.data.shape, 4)


def test_rgba_with_masks():
    da1 = data_array(ndim=2, unit='K')
    da2 = data_array(ndim=2, unit='K', masks=True)
    mapper = ColorMapper()
    assert not np.allclose(mapper.rgba(da1), mapper.rgba(da2))


def test_colorbar_updated_on_rescale():
    da = data_array(ndim=2, unit='K')
    mapper = ColorMapper()
    artist = DummyChild(da)
    key = 'data'
    mapper[key] = artist

    mapper.update(data=da)
    _ = mapper.to_widget()
    old_image = mapper.widget.value
    old_image_array = old_image

    # Update with the same values should not make a new colorbar image
    artist.update(da)
    mapper.update(data=da)
    assert old_image is mapper.widget.value

    # Update with new values should make a new colorbar image
    const = 2.3
    artist.update(da * const)
    mapper.update(data=da * const)
    assert old_image_array != mapper.widget.value


def test_colorbar_does_not_update_on_rescale_if_limits_can_only_grow():
    da = data_array(ndim=2, unit='K')
    mapper = ColorMapper(autoscale='grow')
    artist = DummyChild(da)
    key = 'data'
    mapper[key] = artist

    mapper.update(data=da)
    _ = mapper.to_widget()
    old_image = mapper.widget.value
    old_image_array = old_image

    # Update with the same values should not make a new colorbar image
    artist.update(da)
    mapper.update(data=da)
    assert old_image is mapper.widget.value

    # Update with a smaller range should not make a new colorbar image
    const = 0.8
    artist.update(da * const)
    mapper.update(data=da * const)
    assert old_image is mapper.widget.value

    # Update with new values should make a new colorbar image
    const = 2.3
    artist.update(da * const)
    mapper.update(data=da * const)
    assert old_image_array != mapper.widget.value


def test_colorbar_is_not_created_if_cbar_false():
    mapper = ColorMapper(cbar=False)
    assert mapper.colorbar is None
    assert mapper.cax is None


def test_colorbar_cbar_false_overrides_cax():
    @dataclass
    class Canvas:
        cax: int

    mapper = ColorMapper(cbar=False, canvas=Canvas(cax=0))
    assert mapper.colorbar is None


def test_autoscale_auto_vmin_set():
    da = data_array(ndim=2, unit='K')
    mapper = ColorMapper(autoscale='auto', vmin=-0.5)
    artist = DummyChild(da)
    key = 'data'
    mapper[key] = artist
    mapper.update(data=da)
    assert mapper.vmin == -0.5
    assert mapper.vmax == da.max().value
    # Make sure it handles when da.max() is greater than vmin
    mapper.update(data=da - sc.scalar(5.0, unit='K'))
    assert mapper.vmin == -0.5
    assert mapper.vmin < mapper.vmax


def test_autoscale_auto_vmax_set():
    da = data_array(ndim=2, unit='K')
    mapper = ColorMapper(autoscale='auto', vmax=0.5)
    artist = DummyChild(da)
    key = 'data'
    mapper[key] = artist
    mapper.update(data=da)
    assert mapper.vmax == 0.5
    assert mapper.vmin == da.min().value
    # Make sure it handles when da.min() is greater than vmax
    mapper.update(data=da + sc.scalar(5.0, unit='K'))
    assert mapper.vmax == 0.5
    assert mapper.vmin < mapper.vmax



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import pytest

from plopp import Node
from plopp.backends.matplotlib.interactive import InteractiveFig
from plopp.data.testing import data_array
from plopp.graphics.imageview import ImageView
from plopp.graphics.lineview import LineView


@pytest.mark.usefixtures('_use_ipympl')
def test_logx_1d_toolbar_button():
    da = data_array(ndim=1)
    fig = InteractiveFig(LineView, Node(da), scale={'xx': 'log'})
    assert fig.toolbar['logx'].value


@pytest.mark.usefixtures('_use_ipympl')
def test_logy_1d_toolbar_button():
    da = data_array(ndim=1)
    fig = InteractiveFig(LineView, Node(da), norm='log')
    assert fig.toolbar['logy'].value


@pytest.mark.usefixtures('_use_ipympl')
def test_logxy_2d_toolbar_buttons():
    da = data_array(ndim=2)
    fig = InteractiveFig(ImageView, Node(da), scale={'xx': 'log', 'yy': 'log'})
    assert fig.toolbar['logx'].value
    assert fig.toolbar['logy'].value


@pytest.mark.usefixtures('_use_ipympl')
def test_log_norm_2d_toolbar_button():
    da = data_array(ndim=2)
    fig = InteractiveFig(ImageView, Node(da), norm='log')
    assert fig.toolbar['lognorm'].value



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from collections.abc import Callable
from typing import Any

import ipywidgets as ipw
import scipp as sc

from ..core import node
from ..core.utils import coord_element_to_string
from .box import VBar


class _BaseSliceWidget(VBar):
    def __init__(self, da: sc.DataArray, dims: list[str], range: bool):
        if isinstance(dims, str):
            dims = [dims]
        self._slider_dims = dims
        self.controls = {}
        self.view = None
        children = []

        for dim in self._slider_dims:
            coord = (
                da.coords[dim]
                if dim in da.coords
                else sc.arange(dim, da.sizes[dim], unit=None)
            )
            widget_args = {
                'step': 1,
                'description': dim,
                'min': 0,
                'max': da.sizes[dim] - 1,
                'continuous_update': True,
                'readout': False,
                'layout': {"width": "200px"},
                'style': {'description_width': 'initial'},
            }
            if range:
                slider = ipw.IntRangeSlider(**widget_args)
            else:
                slider = ipw.IntSlider(**widget_args)
            continuous_update = ipw.Checkbox(
                value=True,
                tooltip="Continuous update",
                indent=False,
                layout={"width": "20px"},
            )
            label = ipw.Label(value=coord_element_to_string(coord[dim, 0]))
            ipw.jslink((continuous_update, 'value'), (slider, 'continuous_update'))

            self.controls[dim] = {
                'continuous': continuous_update,
                'slider': slider,
                'label': label,
                'coord': coord,
            }
            slider.observe(self._update_label, names='value')
            children.append(ipw.HBox([continuous_update, slider, label]))

        super().__init__(children)

    def _update_label(self, change: dict[str, Any]):
        """
        Update the readout label with the coordinate value, instead of the integer
        readout index.
        """
        dim = change['owner'].description
        coord = self.controls[dim]['coord'][dim, change['new']]
        self.controls[dim]['label'].value = coord_element_to_string(coord)

    def _plopp_observe_(self, callback: Callable, **kwargs):
        """
        Special method which is used instead of the ``observe`` method of ``ipywidgets``
        because overriding the ``observe`` method of the ``VBox`` causes issues.
        """
        for dim in self.controls:
            self.controls[dim]['slider'].observe(callback, **kwargs)

    @property
    def value(self) -> dict[str, int | tuple[int]]:
        """
        The widget value, as a dict containing the dims as keys and the slider indices
        as values.
        """
        return {dim: self.controls[dim]['slider'].value for dim in self._slider_dims}


class SliceWidget(_BaseSliceWidget):
    """
    Widgets containing a slider for each of the requested dimensions.
    The widget uses the input data array to determine the range each slider should have.
    Each slider also comes with a checkbox to toggle on and off the slider's continuous
    update.

    Parameters
    ----------
    da:
        The input data array.
    dims:
        The dimensions to make sliders for.
    """

    def __init__(self, da: sc.DataArray, dims: list[str]):
        super().__init__(da, dims, range=False)


class RangeSliceWidget(_BaseSliceWidget):
    """
    Widgets containing a slider for each of the requested dimensions.
    The widget uses the input data array to determine the range each slider should have.
    Each slider also comes with a checkbox to toggle on and off the slider's continuous
    update.

    .. versionadded:: 24.04.0

    Parameters
    ----------
    da:
        The input data array.
    dims:
        The dimensions to make sliders for.
    """

    def __init__(self, da: sc.DataArray, dims: list[str]):
        super().__init__(da, dims, range=True)


@node
def slice_dims(data_array: sc.DataArray, slices: dict[str, slice]) -> sc.DataArray:
    """
    Slice the data according to input slices.

    Parameters
    ----------
    data_array:
        The input data array to be sliced.
    slices:
        Dict of slices to apply for each dimension.
    """
    out = data_array
    for dim, sl in slices.items():
        if isinstance(sl, tuple):
            sl = slice(*sl)
        out = out[dim, sl]
    return out
